#!/usr/bin/python

# add_jrc_names.py
# autor Cerny Jan xcerny62@stud.fit.vutbr.cz
# Skript doplni do KB jmena z databaze JRC-Names.
# Skript ma tyto parametry:
# -e, --entities= - jmeno souboru obsahujiciho databazi JRC-Names (povinne)
# -c, --condition - doplni pouze zaznamy, ktere jsou propojeny s JRC-Names
#     aspon pres dve jmena, pricemz jedno musi byt v poli NAME
# -l, --latin - doplni pouze jmena zapsana latinkou
# -v, --verbose - vypise statistiky na standardni chybovy vystup
# Skript nacita KB ze standardniho vstupu.
# Upravenou KB vypisuje na standardni vystup.
# Doporucuji vyuzit presmerovani z/do souboru.
# priklad spusteni: ./add_jrc_names.py -e entities -l -c < KB.all > KB.new
# posledni uprava: 2.dubna 2014

import sys
import re
import getopt

# ziskani parametru:
try:
  opts, args = getopt.getopt(sys.argv[1:], "e:clv", ["entities=","condition","latin","verbose"])
except:
  sys.stderr.write("Spatne parametry, viz README\n")
  exit()

# zpracovani parametru:
file_entities = None
condition_only = False
latin_only = False
verbose = False
for opt, arg in opts:
  if opt in ("-e","--entities"):
    file_entities = arg
  elif opt in ("-c","--condition"):
    condition_only = True
  elif opt in ("-l","--latin"):
    latin_only = True
  elif opt in ("-v","--verbose"):
    verbose = True
    
if file_entities == None:
  sys.stderr.write("Nebylo zadano jmeno souboru obsahujiciho databazi JRC-Names.\n")
  exit()

# regularni vyraz znacici retezce, ktere nejsou v latince
not_latin_pattern = re.compile('[\u02af-\u2000\u206f-\uffff]+')

# nacteni hodnot ze souboru entities: 
entities = open(file_entities, 'r')
idlist = {} # klicem je ID
namelist = {} # klicem je jmeno
for line in entities:
  line_split = line.split('\t')
  if line_split[1] == 'P':  # nacitam jenom osoby, ne organizace
    id = line_split[0]
    name = line_split[3].replace('+',' ').strip()
    if latin_only and re.search(not_latin_pattern, name):
      continue
    if id in idlist:
      idlist[id].append(name)
    else:
      idlist[id] = [name]
    if name in namelist:
      namelist[name].append(id)
    else:
      namelist[name] = [id]
entities.close()

# vytvoreni slovniku z nactenych jmen
# format slovniku: klicem je jmeno, hodnotou je seznam vsech ostatnich jmen
names = {}
for name in namelist:
  names[name] = []
  for id in namelist[name]:
    names[name].extend(idlist[id])

# uz nepotrebuju, tak mazu
idlist.clear()
namelist.clear()

# inicializace pocitadel
added_person = 0
changed_person = 0
added_artist = 0
changed_artist = 0

if condition_only: # s prepinacem --condition
  for line in sys.stdin:
    if line[0] == 'p': #PERSON
      found = False
      line_split = line.split('\t')
      person = line_split[2]
      if line_split[3] and (person in names): # field ALIAS neni prazdny AND jmeno z fieldu NAME je v JRC
        names_to_add = list(names[person])
        names_to_add.remove(person)
        aliases = line_split[3].split('|')
        for x in aliases:
          if x in names_to_add:
            found = True
            break
        if found:
          for x in aliases:
            try:
              names_to_add.remove(x)
            except:
              pass
          added_person += len(names_to_add)
          changed_person +=1
          line_split[3] =  '|'.join(aliases+names_to_add)
          line = '\t'.join(line_split)
    elif line[0] == 'a': #ARTIST
      found = False
      line_split = line.split('\t')
      person = line_split[2]
      if (line_split[4] or line_split[3]) and (person in names):
        names_to_add = list(names[person])
        names_to_add.remove(person)
        aliases = []
        if line_split[4]:
          aliases.extend(line_split[4].split('|'))
        if line_split[3]:
          aliases.append(line_split[3])
        for x in aliases:
          if x in names_to_add:
            found = True
            break
        if found:
          aliases.remove(line_split[3])
          for x in aliases:
            try:
              names_to_add.remove(x)
            except:
              pass
          added_artist += len(names_to_add)
          changed_artist += 1
          line_split[4] =  '|'.join(aliases+names_to_add)
          line = '\t'.join(line_split)
    sys.stdout.write(line)
else: # bez prepinace --condition
  for line in sys.stdin:
    if line[0] == 'p': #PERSON
      found = False
      line_split = line.split('\t')
      person = line_split[2]
      if line_split[3]: # jsou nejake aliasy ?
        aliases = line_split[3].split('|')
      else:
        aliases = []
      names_in_kb = [person] + aliases 
      for x in names_in_kb:
        if x in names:
          found = x
          break
      if found:
        names_to_add = list(names[found])
        for x in names_in_kb:
          try:
            names_to_add.remove(x)
          except:
            pass
        added_person += len(names_to_add)
        changed_person += 1
        line_split[3] =  '|'.join(aliases+names_to_add)
        line = '\t'.join(line_split)
    elif line[0] == 'a': #ARTIST
      found = False
      line_split = line.split('\t')
      person = line_split[2] # O3 DISPLAY TERM
      if line_split[4]: # 05 OTHER TERM (MULTIPLE VALUES)
        aliases = line_split[4].split('|')
      else:
        aliases = []
      names_in_kb = [person] + aliases
      if line_split[3]: # 04 PREFFERED TERM
        names_in_kb.append(line_split[3])
      for x in names_in_kb:
        if x in names:
          found = x
          break
      if found:
        names_to_add = list(names[found])
        for x in names_in_kb:
          try:
            names_to_add.remove(x)
          except:
            pass
        added_artist += len(names_to_add)
        changed_artist += 1
        line_split[4] =  '|'.join(aliases+names_to_add)
        line = '\t'.join(line_split)
    sys.stdout.write(line)

# vypis informaci
if verbose: 
  sys.stderr.write("Zmeneno person: {0}\n".format(changed_person))
  sys.stderr.write("Zmeneno artist: {0}\n".format(changed_artist))
  sys.stderr.write("Zmeneno celkem: {0}\n".format(changed_artist+changed_person))
  sys.stderr.write("Pridano k person: {0}\n".format(added_person))
  sys.stderr.write("Pridano k artist: {0}\n".format(added_artist))
  sys.stderr.write("Pridano celkem: {0}\n".format(added_artist + added_person))
