#!/bin/bash
#
#Copyright 2015 Brno University of Technology
#
#Licensed under the Apache License, Version 2.0 (the ""License"");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an ""AS IS"" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

# Author: Martin Graca, xgraca00@stud.fit.vutbr.cz
#
# Description: script for testing start.sh - check correct format of required files,
#		 existing of files, rights; comparing olds and news created automats
#		 testing ner.py and figa  
#last change: 26. 2. 2015

########################################################################################
##	list of file which are use to create KB

prefix="/mnt/minerva1/nlp/projects/";
prefix2="/mnt/minerva1/nlp/datasets/art/";
types=("PERSONS" "ARTISTS" "LOCATIONS" "ARTWORKS" "MUSEUMS" "MYTHOLOGY" "ARTIST_FAMILIES" "ARTIST_GROUP_OR_COLLECTIVE") 
fbtypes=( "artists" "artworks" "persons" "events" "visual_art_forms" "visual_art_mediums" "museums" "art_period_movements" "visual_art_genres" "nationalities" )
statistics=( "locations" "artists" "artworks" "books" "museums" "persons" "writers" )
others=( "decipher_ner/ULAN/ulan_rel_13/KB.artists" "decipher_wikipedia/wiki_template/artists_extended" \
	"decipher_dbpedia/extraction_results/v39/v39_artwork_finall.tsv" "decipher_dbpedia/extraction_results/v39/v39_location_finall.tsv" \
	"decipher_dbpedia/extraction_results/v39/v39_museum_finall.tsv" "decipher_ner/ULAN/ulan_rel_13/KB.corporations" \
	"extrakce_z_wikipedie/xgraca00/MythologyKB.txt" "extrakce_z_wikipedie/xdosta40/final/finalFamilies.xml" \
	"extrakce_z_wikipedie/xdosta40/final/finalGroupsAndCollectives.xml" \
	"decipher_geonames/geonames.locations" "decipher_geonames/geonames.museums" "decipher_wikipedia/Ludia_Wikipedia/outputs/categories_death_birth_data.csv" \
	"ie_foreign_wikipedia/xbajan01/wiki/output/es_only_data.tsv" "ie_foreign_wikipedia/xklima22/wiki/outputs/de-wiki_only_people_data.tsv" )
datasets=( "artbiogs/artbiogs.tsv" "bbc/bbc.tsv" "council/council.tsv" "distinguishedwomen/distinguishedwomen.tsv" \
	"artists2artists/artists2artists.tsv" "the-artists/the-artists.tsv" "nationalgallery/nationalgallery.tsv" \
	"globalgallery/globalgallery.tsv" "wikipaint/final_data/wikipaint_artist.tsv" "davisart/davisart.tsv" "apr/apr.tsv" "rkd/rkd.tsv" \
	"open/open.tsv" )
#########################################################################################
check_number_of_columns(){ #checking correct number of columns in KB files

	find KnowBase -name '*.fields' | xargs wc -l | sed s/^\ \*//g | sed '$d' > fields.txt
	python MatchKBAndFields.py | sed '/^$/d'> KBAndFields.txt
	printf '%s' ' /mnt/minerva1/nlp/projects/entity_kb_czech2/xviden04/artists/artistsCSwiki.tsv 27 KnowBase/artists/ARTISTS.fields' >> KBAndFields.txt
        printf '%s' ' /mnt/minerva1/nlp/projects/ie_foreign_wikipedia/xbajan01/wiki/output/es_only_data.tsv 18 KnowBase/persons/PERSONS.fields' >> KBAndFields.txt
        printf '%s' ' /mnt/minerva1/nlp/projects/ie_foreign_wikipedia/xklima22/wiki/outputs/de-wiki_only_people_data.tsv 18 KnowBase/personss/PERSONS.fields' >> KBAndFields.txt
	rm fields.txt
	while read line
	do
		file=$(echo $line | awk '{print $1}' | awk -F ':' '{print $1}') # test file
		Columns=$(echo $line | awk '{print $2}') #right number of Columns 
		numberOfC=$(awk -F '\t' '{ print NF }' "$file" | sort -nu | uniq) # current number of Columns 
		if [[ $numberOfC -eq $Columns ]]; then # same number of columns
			echo "OK $file"
			true
		else 
			echo "bad number of columns in file: $file: $numberOfC correct: $Columns"
		fi
		
	done < KBAndFields.txt
	rm KBAndFields.txt

}
#######################################################################################		
check_exists(){ # check existing of files
	for type in "${datasets[@]}"; do
		if [ -e ""$prefix2""$type"" ]; then # if file exists
			true
		else
			echo "file "$prefix2""$type" doesn't exist"
		fi
	done
	for type in "${types[@]}"; do
		if [ -e ""$prefix"decipher_ner/KnowBase/"$type"" ]; then # if file exists
			true
		else
			echo "file "$prefix"decipher_ner/KnowBase/"$type" doesn't exist"
		fi
	done

	for type in "${others[@]}"; do
		if [ -e ""$prefix""$type"" ]; then # exists
			true
		else
			echo "file "$prefix""$type" doesn't exist"
		fi
	done
	for type in "${fbtypes[@]}"; do
		if [ -e ""$prefix"decipher_freebase/data/freebase."$type"" ]; then # exists
			true
		else
			echo "file "$prefix"decipher_freebase/data/freebase."$type" doesn't exist"
		fi
	done
}
#######################################################################################
check_rights(){ # of using files
		files=( $1* )  # files in directory
		for file in "${files[@]}"; do
			perms=$(stat -c %a "$file")
			if [ -f "$file" ]; then
				if [ "$perms" = "660" ]; then # bad rights
					if [[ ${file: -3} = ".py" ]]; then
						echo "$file has $perms permisions" 
					elif [[ ${file: -3} = ".sh" ]]; then
						echo "$file has $perms permisions"
					else
						true
					fi
				fi
				if [ -r "$file" ]; then # read rights
					true
				elif [ -w "$file" ]; then # wright rights
					true
				elif [ -x "$file" ]; then
					true
				else
					echo "$file has not permissions"
				fi
			elif [ -d "$file" ]; then
				#"$file is a directory"
				check_rights "$file/"
			else
				true
			fi
		done
}
#######################################################################################
check_fsa_kb(){ # comparing olds and news automats
	ssh athena3.fit.vutbr.cz "ls -l /mnt/athena3/kb | sed 1d" | sed -n '/.*.fsa.*\|.*KB.*.all.*/p'  > old.txt # list of old fsa files and KB
	cat old.txt | awk -v OFS=';' '{ print $5,$9 }' > old_kbfsa_sizes.txt
	rm old.txt
	prefix="figa/";
	cat old_kbfsa_sizes.txt | while read line; do
		set -- "$line"
		IFS=";"; declare -a Array=($*)
		if [[ -e "${Array[1]}" || -e "$prefix${Array[1]}" ]]; then
			# get size of file
			if [[ "${Array[1]}" =~ .*.fsa.* ]]; then
				newfilesize=$(stat -c%s "$prefix${Array[1]}")
			else
				newfilesize=$(stat -c%s "${Array[1]}")
        	       	fi
			rozdil=$(echo "$newfilesize/${Array[0]}" | bc -l) # size difference   
               		introzdil=$(( newfilesize / Array[0] )) # int size difference
			hranice=$(echo "105/100" | bc -l) # high boundary
			dolnihranice=$(echo "90/100" | bc -l) # low boundary
			if [ "$introzdil" = "1" ]; then
				bool=$(echo $rozdil'>'$hranice | bc -l) #check top boundary
				if [ "$bool" = "1" ]; then
					echo "new ${Array[1]} is 5% and more bigger then old version"
				fi
			else
				bool2=$(echo $rozdil'<'$dolnihranice | bc -l) # check bottom boundary
				if [ "$bool2" = "1" ]; then
					echo "new ${Array[1]} is 5% and more smaller than old version"
				fi
			fi
		else
			echo " new ${Array[1]} wasn't created"
		fi
	done
rm old_kbfsa_sizes.txt
}
#######################################################################################
check_duplicite_lines(){ 
	# in all KB files
	ok=false;
	ok2=false;
	for type in "${datasets[@]}"; do
		 # save duplicite lines to file
		sort "$prefix2""$type" | uniq --count | sed -n '/^      1.*/!p' > dupl.txt #check whole line
		if [[ -s dupl.txt ]]; then # if file is not empty
			 echo "file "$prefix2""$type" contains duplicite lines:"
			cat dupl.txt # print founded duplicite lines
			ok=true;
		fi
		if [[ -s bad_separ.txt ]]; then
			echo "file "$prefix2""$type" has bad separators:"
			cat bad_separ.txt # print line with bad separators
			ok2=true;
                elif [[ $ok = false && $ok2 = false ]]; then # no duplicite lines
                        echo "OK $prefix2$type" 
		else 
			true
                fi
		ok=false;
		ok2=false;
        done
	i=0;
	ok=false;
	ok2=false;
	#same as first for
        for type in "${types[@]}"; do
                sort ""$prefix"decipher_ner/KnowBase/"$type"" | uniq --count | sed -n '/^      1.*/!p' > dupl.txt  

		if [[ -s dupl.txt ]]; then
                        echo "file $prefix"decipher_ner/KnowBase/"$type"" contains duplicite lines:"
			cat dupl.txt
			ok=true;
		fi
		if [[ -s bad_separ.txt ]]; then
                        echo "file ""$prefix"decipher_ner/KnowBase/"$type"" has bad separators:"
                        cat bad_separ.txt
                        ok2=true;
                elif [[ $ok = false && $ok2 = false ]]; then # no duplicite lines
                        echo "OK "$prefix"decipher_ner/KnowBase/"$type""
		else
			true
		fi
		i=$((i+1));
		ok=ok2=false;
        done
	#same as first for
        for type in "${others[@]}"; do
                sort "$prefix""$type" | uniq --count | sed -n '/^      1.*/!p' > dupl.txt
                if [[ -s dupl.txt ]]; then
                        echo "file "$prefix""$type" contains duplicite lines:"
			cat dupl.txt
                else
                        echo "OK "$prefix""$type"" 
                fi
        done
	#same as first for
        for type in "${fbtypes[@]}"; do
                sort ""$prefix"decipher_freebase/data/freebase."$type"" | uniq --count | sed -n '/^      1.*/!p' > dupl.txt  
                if [[ -s dupl.txt ]]; then
                        echo "file ""$prefix"decipher_freebase/data/freebase."$type"" contains duplicite lines:"
			cat dupl.txt
                else # no duplicite lines
                        echo "OK "$prefix"decipher_freebase/data/freebase."$type""
                fi
        done
	rm dupl.txt
}
check_ner_out(){
	# compare hand anotation with Ner output
	
	# neccesarry for ner.py
	./deleteKB.sh
	./downloadKB.sh
	make
	echo "comparing Paul_Kane:"
	./ner.py -f data/input/Paul_Kane > Paul_Kane.txt
	python CheckNerOut.py -o Paul_Kane.txt -a /mnt/minerva1/nlp/projects/decipher_ner/xmagam00/annotation/annotation_PaulKane_xmagam00_v02.tsv
	echo "comparing travelling_artist_part2:"
	./ner.py -f data/input/traveling_artist_part_2 > Trav_Art2.txt
	python CheckNerOut.py -o Trav_Art2.txt -a /mnt/minerva1/nlp/projects/decipher_ner/xmagam00/annotation/annotation_travelingArtistPart2_xmagam00_v02.tsv	
	echo "comparing travelling_artist_part3:"
        ./ner.py -f data/input/traveling_artist_part_3 > Trav_Art3.txt
        python CheckNerOut.py -o Trav_Art3.txt -a /mnt/minerva1/nlp/projects/decipher_ner/xmagam00/annotation/annotation_travelingArtistPart3_xmagam00_v02.tsv
	echo "comparing Dossier_and_OS_texts.docx:"
        ./ner.py -f data/input/Dossier_and_OS_texts.docx > Doss_And_Os_Texts.txt
        python CheckNerOut.py -o Doss_And_Os_Texts.txt -a /mnt/minerva1/nlp/projects/decipher_ner/xmagam00/annotation/annotation_DossierAndOsTexts_xmagam00_v02.tsv

	# deleting temporary file
	rm Paul_Kane.txt
	rm Trav_Art2.txt
	rm Trav_Art3.txt
	rm Doss_And_Os_Texts.txt

}
#######################################################################################

testing=false
run=false
help_h_param=false
figa=false
ner=false
badarg=false
help_param=false
checknerout=false
# arguments
if [ "$1" = "-help" ]; then
	help_param=true
fi
if [ "$1" = "--help" ]; then
	help_param=true
fi


while getopts :trhfnc param
do	
	case "$param" in
	t) testing=true;;
	r) run=true;;
	h) help_h_param=true;;
	f) figa=true;;
	n) ner=true;;
	c) checknerout=true;;
	*) help_h_param=true; badarg=true  >&2;;
	esac 
done
# no arguments
if [ "$checknerout" = false -a "$help_h_param" = false -a "$run" = false -a "$testing" = false -a "$figa" = false -a "$ner" = false ] ; then
	echo "please enter arguments or run ./TestAndRunStart.sh -h to show help"
fi
if [ "$help_h_param" = true ] ; then #print help
	if [ "$badarg" = true -a "$help_param" = false ] ; then
		echo "Bad arguments"
	fi
echo "run examples:" 
echo "		 ./TestAndRunStart.sh -t -> run test existing of file, rights and test format of KB files" 
echo "		 ./TestAndRunStart.sh -r -> run start.sh "
echo "		 ./TestAndRunStart.sh -f -> run testing figa"
echo "		 ./TestAndRunStart.sh -n -> run testing ner"
echo "		 ./TestAndRunStart.sh -h -> print help of script"
echo "		 ./TestAndRunStart.sh -c -> run script CheckNerOut.py - test ner output with anotation"
fi

# testing part
if [ "$testing" = true -a "$badarg" = false ] ; then
		echo "----------testing missing files----------------------------"
check_exists
		echo "----------testing files permissions------------------------"
check_rights
		echo "----------testing number of columns------------------------"
check_number_of_columns
                echo "----------testing duplicite lines in KB files--------------"
check_duplicite_lines
fi
if [ "$checknerout" = true ]; then
		echo "----------comparing anotation with NER Output--------------"
check_ner_out

fi
# running part
if [ "$run" = true -a "$badarg" = false ] ; then
	# catch errors to the file
	./start.sh 2> errors.txt
	errorcounter=$(cat errors.txt | grep -c "^[^ ]*Error:") #Errors
	if [ $errorcounter == 0 ]; then # no errors
		echo "----------script ends with ${errorcounter} error-------------------------"
	elif [ $errorcounter == 1 ]; then # one error
		echo "----------script start.sh ends with ${errorcounter} error:------------------------" >&2
		cat errors.txt | sed -n '/Traceback/,/Error:/p' >&2
	else # 2 and more err
		echo "----------script start.sh ends with ${errorcounter} errors:-----------------------" >&2
		cat errors.txt | sed -n '/Traceback/,/Error:/p' >&2
	fi
	grep 'soubor neexistuje' errors.txt > missingfiles.txt
	missingfiles=$(egrep -o '„[^ ]*:' missingfiles.txt)
	numberOfLines=$(wc -l missingfiles.txt | awk '{ print $1 }')
	if [[ -z ${missingfiles} ]]; then
		true
	else		
		echo "----------${numberOfLines} missing files:---------------------------------"
		echo "${missingfiles}"
	fi
	sed '/soubor neexistuje/d' errors.txt | sed '/Traceback/,/Error:/d' > nexterr.txt 
	if [[ -s nexterr.txt ]]; then
		echo "---------next warnings  start.sh:------------------------------------" >&2
        	sed '/soubor neexistuje/d' errors.txt | sed '/Traceback/,/Error:/d' >&2
	fi
	rm nexterr.txt
	rm missingfiles.txt
	rm errors.txt

		echo "---------testing kb and automats---------------------------"
check_fsa_kb
fi
if [ "$ner" = true -a "$badarg" = false ] ; then
		echo "---------testing ner.py------------------------------------"
# testing ner on basic data
./ner.py -f data/input/example_input2 > output.txt 2>> kberrors.txt
grep -v '^DEBUG.*$' kberrors.txt > greped.txt
rm kberrors.txt
cat greped.txt
if [ -s "greped.txt" ]; then
	echo "1. test: ./ner.py -f data/input/example_input2 failed:"
	cat greped.txt
else
	echo "1. test: ./ner.py -f data/input/example_input2 OK"
fi
rm greped.txt
./ner.py -a <<<"Paris 1984" > output.txt 2>> kberrors.txt
grep -v '^DEBUG.*$' kberrors.txt > greped.txt
rm kberrors.txt
if [ -s "greped.txt" ]; then
        echo "2. test: ./ner.py -a <<<"Paris 1984" failed:"
	cat greped.txt
else
        echo "2. test: ./ner.py -a <<<"Paris 1984" OK"
fi
rm greped.txt
./ner.py -f data/input/Paul_Kane > output.txt 2>> kberrors.txt
grep -v '^DEBUG.*$' kberrors.txt > greped.txt
rm kberrors.txt
if [ -s "greped.txt" ]; then
	echo "3. test: ./ner.py -f data/input/Paul_Kane failed:"
	cat greped.txt
else
	echo "3. test: ./ner.py -f data/input/Paul_Kane OK"
fi
rm greped.txt
./ner.py -f data/input/The_Four_Seasons.txt > output.txt 2>> kberrors.txt
grep -v '^DEBUG.*$' kberrors.txt > greped.txt
rm kberrors.txt
if [ -s "greped.txt" ]; then
	echo "4. test: ./ner.py -f data/input/The_Four_Seasons.txt failed:"
	cat greped.txt
else
	echo "4. test: ./ner.py -f data/input/The_Four_Seasons.txt OK"
fi
rm greped.txt
./ner.py -f data/input/traveling_artist_part_2 > output.txt 2>> kberrors.txt
grep -v '^DEBUG.*$' kberrors.txt > greped.txt
rm kberrors.txt
if [ -s "greped.txt" ]; then
	echo "5. test: ./ner.py -f data/input/traveling_artist_part_2 failed:"
	cat greped.txt
else
        echo "5. test: ./ner.py -f data/input/traveling_artist_part_2 OK"
fi
rm greped.txt
./ner.py -f data/input/traveling_artist_part_3 > output.txt 2>> kberrors.txt
grep -v '^DEBUG.*$' kberrors.txt > greped.txt
rm kberrors.txt
if [ -s "greped.txt" ]; then
        echo "6. test: ./ner.py -f data/input/traveling_artist_part_3 failed:"
	cat greped.txt
else
        echo "6. test: ./ner.py -f data/input/traveling_artist_part_3 OK"
fi
rm output.txt
rm greped.txt
#####################################################################
# Simple test of some NER properties.
# author: Jan Černý
# last change: 15. July 2014


echo "Common test (without using parameters)"
echo "=================================================================="
tests=("Paris" "Brittany" "Agatha Christie" "Bacchus" "Buenos Aires" "Second world war")
for name in "${tests[@]}"
do
    echo "$name"
    linenum=`echo "$name" | ./ner.py | cut -f5`
    cat -n KB.all | head -$linenum | tail -1 | cut -f-4
done
echo ""

echo "Test with parameter -l --- conversion input to small letters"
echo "=================================================================="
lowertests=("Leonardo DA VINCI" "LEONARDO DA VINCI" "Claude MONET" "DICK Francis" "VASCO da GAMA" "Ban KI-Moon")
for name in "${lowertests[@]}"
do
    echo "$name"
    linenum=`echo "$name" | ./ner.py -l | cut -f5`
    cat -n KB.all | head -$linenum | tail -1 | cut -f-4
done
echo ""

echo "Test with parameter -r --- errors in diacritics in input"
echo "=================================================================="
accenttests=("Paul Cèzanne" "Hónóré dé Bálzác" "Louis de Funés" "Chalôns-sur-Marne" "Bésäncon" "Strakoňice")
for name in "${accenttests[@]}"
do
    echo "$name"
    linenum=`echo "$name" | ./ner.py -r | cut -f5`
    cat -n KB.all | head -$linenum | tail -1 | cut -f-4
done
echo ""

echo "Test with combination parameters -l a -r (really bad input text)"
echo "=================================================================="
kombitests=("Georges BRASSÉNS" "ORLICKE hory" "TYNEČ nad labem" )
for name in "${kombitests[@]}"
do
    echo "$name"
    linenum=`echo "$name" | ./ner.py -lr | cut -f5`
    cat -n KB.all | head -$linenum | tail -1 | cut -f-4
done
fi
if [ "$figa" = true -a "$badarg" = false ] ; then
cd figa
# testing figa on basic dat
	echo "---------testing figa--------------------------------------"
	if [ -e "automata.fsa" ]; then
		./figav08 -d automata.fsa <<< "A Alewijn" > output.txt 2>> fsaerrors.txt
		if [ -s "fsaerrors.txt" ]; then
	        	echo "1. test: ./figav08 -d automata.fsa <<< "A Alewijn" failed:"
        		cat fsaerrors.txt
		else # test OK
	        	echo "1. test: ./figav08 -d automata.fsa <<< "A Alewijn" OK"
		fi
		rm output.txt
		rm fsaerrors.txt
		./figav08 -d automata.fsa < ../data/input/example_input2 > output.txt 2>> fsaerrors.txt
        	if [ -s "fsaerrors.txt" ]; then
                	echo "2. test: ./figav08 -d automata.fsa < ../data/input/example_input2 failed:"
               	 	cat fsaerrors.txt
        	else # test OK
                	echo "2. test: ./figav08 -d automata.fsa < ../data/input/example_input2 OK"
        	fi
        	rm output.txt
        	rm fsaerrors.txt

	else
		echo "can't test figa"
		echo "file automata.fsa doesn't exist"
	fi
fi
if [ "$help_h_param" = false -a "$badarg" = false ]; then
	echo "testing ended"
fi
