#!/bin/sh

# Author: Lubomir Otrusina, iotrusina@fit.vutbr.cz

export LC_ALL="C"

# default values
KB="KBstatsMetrics.all"
LOWERCASE=false
SPELLCHECKING=false
URI=false

#=====================================================================
# nastavovani parametru prikazove radky

usage()
{
    echo "Usage: create_fsa.sh [-h] [-l|-s|-u] --knowledge-base=KBstatsMetrics.all"
    echo ""
    echo "\t-h --help"
    echo "\t-l --lowercase"
    echo "\t-s --spellchecking"
    echo "\t-u --uri"
    echo "\t-k --knowledge-base=$KB"
    echo ""
}

while [ "$1" != "" ]; do
    PARAM=`echo $1 | awk -F= '{print $1}'`
    VALUE=`echo $1 | awk -F= '{print $2}'`
    case $PARAM in
        -h | --help)
            usage
            exit
            ;;
        -l | --lowercase)
            LOWERCASE=true
            ;;
        -s | --spellchecking)
            SPELLCHECKING=true
            ;;
        -u | --uri)
            URI=true
            ;;
        -k | --knowledge-base)
            KB=$VALUE
            ;;
        *)
            echo "ERROR: unknown parameter \"$PARAM\""
            usage
            exit 1
            ;;
    esac
    shift
done

#=====================================================================
# vytvoreni seznamu klicu entit v KB

if $LOWERCASE ; then
  python KB2namelist.py -l < "$KB" | tr -s ' ' > intext
elif $URI ; then
  python KB2namelist.py -u < "$KB" > intext
else
  python KB2namelist.py < "$KB" | tr -s ' ' > intext
fi

#=====================================================================
# pridani fragmentu jmen a prijmeni entit

if ! $URI ; then
  python get_names.py > names
  sort -u < names | sed 's/$/\tN/' > fragments
  cat fragments >> intext
fi

#=====================================================================
# uprava stoplistu (kapitalizace a razeni)

python get_morphological_forms.py < stop_list | sort -u > stop_list.var
cp stop_list.var stop_list.all
sed -e 's/\b\(.\)/\u\1/g' < stop_list.var >> stop_list.all
tr 'a-z' 'A-Z' < stop_list.var >> stop_list.all
tr 'A-Z' 'a-z' < stop_list.var >> stop_list.all
sort -u stop_list.all > stop_list.all.sorted

#=====================================================================
# parsovanie confidence hodnot do samostatneho suboru
# redukcia duplicit, abecedne zoradenie entit
# odstranovani slov ze stop listu

if ! $URI ; then
  awk '{print $(NF)}' < "$KB" > KB_confidence
  python uniq_namelist.py -s "KB_confidence" < intext | tr '\t' '|' > namelist.UTF-8
else
  python uniq_namelist.py < intext | tr '\t' '|' > namelist.UTF-8
fi

#=====================================================================
# prevod z UTF-8 znaku na ASCII reprezentaci

./utf2symbols < namelist.UTF-8 > namelist.ASCII

#=====================================================================
# prevod cisel riadkov z ASCII do binarnej podoby

if ! $SPELLCHECKING ; then
  ./str2bin < namelist.ASCII | LANG=C sort -u > namelist.ASCII.bin
fi

#=====================================================================
# vytvoreni konecneho automatu

if $LOWERCASE ; then
  fsa/fsa_build < namelist.ASCII.bin > ../automata-lower.fsa
elif $URI ; then
  fsa/fsa_build < namelist.ASCII.bin > ../automata-uri.fsa
elif $SPELLCHECKING ; then
  cat namelist.ASCII | cut -d'|' -f1 | LANG=C sort -u > namelist.spell
  fsa/fsa_build < namelist.spell > ../automata-spell.fsa
else
  fsa/fsa_build < namelist.ASCII.bin > ../automata.fsa
fi

#=====================================================================
# smazani pomocnych souboru

rm -f names
rm -f fragments
rm -f intext
rm -f stop_list.all stop_list.var stop_list.all.sorted
rm -f namelist.UTF-8
rm -f namelist.ASCII
rm -f namelist.spell
rm -f namelist.ASCII.bin
rm -f KB_confidence
