/*
Copyright 2014 Brno University of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

//=====================================================//
// Soubor: figa.cc v0.8                                //
// Autor: Marek Visnovsky,  xvisno00@stud.fit.vutbr.cz //
// Upravil: Karel Brezina,  xbrezi13@stud.fit.vutbr.cz //
// Upravil: Peter Hostacny, xhosta03@stud.fit.vutbr.cz //
//=====================================================//

#include <ctype.h>
#include <cmath>
#include "figa.h"
#include <list>
#include <vector>

static const int	MAX_LINE_LEN = 512;

// Trida marker

//================================//
// Nazev: initList                //
// Trida: marker                  //
// Popis: pocatecni inicializace  //
//================================//
void marker::initList()
{
	ulist.First = NULL;
	ulist.Act = NULL;
	ulist.Move = NULL;
} // konec initList

//=================================================//
// Nazev: insertItem                               //
// Trida: marker                                   //
// Popis: provede vlozeni nove polozky do seznamu  //
//=================================================//
void marker::insertItem(unsigned int offset, unsigned int value, unsigned long int char_value)
{
	TListptr* item;
	item = (TListptr*)malloc(sizeof(struct ListItem));
	item->char_value = char_value;

	if (ulist.First == NULL) { // je pridavana polozka prvni?
		ulist.First = item;
		ulist.Act = item;
		ulist.Move = item;
		item->count_sub = value;
		item->offset_id = offset;
		item->next = NULL;
	}
	else {
		ulist.Move->next = item;
		item->count_sub = ulist.Move->count_sub + value;
		item->offset_id = offset;
		item->next = NULL;
		ulist.Move = item;
	}

} // konec insertItem

//===========================================//
// Nazev: deleteItems                        //
// Trida: marker                             //
// Popis: vymaze vsechny polozky v seznamu   //
//===========================================//
void marker::deleteItems()
{
	TListptr* help;
	while (ulist.First != NULL) { // je posledni polozka?
		help = ulist.First;
		ulist.First = ulist.First->next;
		free(help);
	}
	ulist.Act = NULL;
	ulist.Move = NULL;
} // konec marker

// Trida data_proc

//==========================================//
// Nazev: get_line                          //
// Trida: data_proc                         //
// Popis: nacitaci sekvence pro hash cisla  //
//==========================================//
void data_proc::save_line(unsigned long int radek, bool end)
{
	std::stringstream string_tmp;
	if (end) { // je cislo radku posledni?
		string_tmp << radek;
		radek_line.append(string_tmp.str());
	}
	else {
		string_tmp << radek << ';';
		radek_line.append(string_tmp.str());
	}
} // konec data_proc::get_line

//=======================================//
// Nazev: get_data                       //
// Trida: data_proc                      //
// Popis: vypis dat o entite do stringu  //
//=======================================//
void data_proc::get_data(Tmatch match, int choice)
{
	std::stringstream string_tmp;
	if (choice == AUTOCOMPLETE)
		string_tmp << match.line_name << '\t' << match.line_num << endl;
	else {
		string_tmp << match.line_num << '\t' << match.first_index << '\t' << match.last_index << '\t' << match.line_name;
		if (choice == FIGA)
			string_tmp << '\t' << "F" << std::endl;
		else
			string_tmp << '\t' << "S" << std::endl;
	}
	results.append(string_tmp.str()); // ulozeni nactenych dat
} // konec data_proc::get_data

// Trida marker

//=================================================//
// Nazev: to_skip                                  //
// Trida: marker                                   //
// Popis: urcuje, ktere znaky se maji preskakovat  //
//=================================================//
bool marker::to_skip(unsigned long int znak, int offset)
{
	// kontrola ci ';' nie je ukoncovaci znak jedneho z utf-8 znakov
	if (offset != -1 && znak == ';' && ulist.Move != NULL) {
		for (TListptr *utf = ulist.First; utf != NULL; utf = utf->next) {
			if (offset == utf->offset_id) {
				return false;	// false ak je ';' ukoncenie utf-8 znaku
			}
		}
	}

	if (znak == ';' && ulist.Move != NULL && line_index == ulist.Move->offset_id) {
		return false;	// false ak je ';' ukoncenie utf-8 znaku
	}

	if (strchr(delimiters, znak))
		return true;

	switch (znak) {
	//case ' ':
	//case '\"':
	//case '\'':
	//case '.':
	//case ',':
	//case '\t':
	//case '\0':
	case 0xE2809D: // right double quotation mark
	case 0xE28093: // en dash
	case 0xE2809C: // left double quotation mark
		return true; // preskocit
	default:
		return false; // nepreskakovat
	}
} // konec marker::to_skip

//========================================================//
// Nazev: set_index                                       //
// Trida: marker                                          //
// Popis: nastavi index udavajici pozici v texte na nulu  //
//========================================================//
void marker::set_index()
{
	index = -1;
	first = 0;
	line_index = 0;
	a = ' ';
	b = ' ';
	index_global = 0;                // reset
	line.clear();
	first_par = true;
	was_end = false;

	initList(); // inicializace seznamu pro index utf-8 znaku
} // konec marker::set_index

//==================================================//
// Nazev: dec2hex                                   //
// Trida: marker                                    //
// Popis: prevede decimalni cislo na hexadecimalni  //
//==================================================//
void marker::dec2hex(char *ext, unsigned int znak)
{
	unsigned int kolik = 0;
	unsigned int hex = 0;
	for (int i = 1; i > -1; i--) {
		kolik = znak / 16;
		hex = znak % 16;
		znak = kolik;
		if (hex < 10) { // rozsah (0-9)
			ext[i] = hex + '0';
		}
		else { // rozsah (A-F)
			ext[i] = hex + 'A' - 10;
		}
	}
} // konec marker::dec2hex

//==================================================//
// Nazev: hex2dec                                   //
// Trida: marker                                    //
// Popis: prevede hexadecimalni znak na cislo	    //
//==================================================//
int marker::hex2dec(char ext)
{

	if (ext >= 'A' && ext <= 'F') {
		return ext - 'A' + 10;
	}
	else {
		return ext - '0';
	}
} // konec marker::hex2dec

//===================================================//
// Nazev: utf2symbols                                //
// Trida: marker                                     //
// Popis: prevede utf-8 znaky na ascii reprezentaci  //
//===================================================//
std::string marker::utf2symbols(std::string line)
{
	unsigned int count = 0;
	unsigned int index = 0;
	unsigned int neco = line[0];
	unsigned int pocet = 0;

	char pole[2];
	char* buf;
	buf = &pole[0];

	char buffer;
	std::string out = "";
	unsigned long int char_value = 0;

	while (neco != '\0') {
		char_value = 0;

		if ((neco >= 0) && (neco < 128)) { // je znak v rozsahu
			buffer = neco;                  // zakladni ASCII sady
			out.append(1, buffer); // pridani do vystupniho retezce
			count++;
			index++;
			neco = line[count]; // nacteni dalsiho znaku
		}
		else {
			out.append("&#x"); // predpona utf-8 znaku
			index += 3;
			pocet = 2;

			for (int i = 6; i > 0; i--) { // na kolik bytu je utf-8 znak
				if (!((neco >> i) & 1)) {
					pocet = 7 - i;
					break;
				}
			}

			for (int i = 0; i < pocet; i++) {

				dec2hex(buf, neco); // prevod na decimalni hodnotu
				for (int j = 0; j < 2; j++) {
					out.append(1, pole[j]); // pridani do vystupniho retezce
					char_value += hex2dec(pole[j]) * pow(16.0, (pocet * 2 - 1) - i * 2 - j);
					index++;
				}

				count++;
				neco = line[count]; // nacteni dalsi casti utf-8 znaku
			}

			out.append(";"); // pripona utf-8 znaku
			index++;
			if (offset_in_bytes)
				insertItem(index, pocet + 4, char_value); // zaindexovani konce utf-8
			else
				insertItem(index, (pocet * 2) + 3, char_value); // zaindexovani konce utf-8
		}                                  // znaku
	}

	ulist.Move = ulist.First; // posunutí indexu

	return out;
} // konec utf2symbols

//=====================================//
// Nazev: next_char                    //
// Trida: marker                       //
// Popis: nacita dalsi znak ze vstupu  //
//=====================================//
void marker::next_char()
{
	bool delimiter = true; // indikacia ci bol najdeny ukoncovaci znak ohranicenia

	// na konci radku nacitam dalsi
	if (line.empty() || line_index == line.length()+1) { // konec radku?
		if (ulist.Move != NULL) { // zacatek nacitani?
			index_global += line.length() + 1 - ulist.Move->count_sub;
		} // index konce predchoziho radku
		else {
			index_global = 0; // zacatek
		}
		deleteItems(); // reset seznamu indexu utf-8 znaku
		insertItem(1,0,0); // nastaveni pocatecni hodnoty
		line_index = 0; // reset
		getline(*input, line); // nacteni noveho radku
		line = utf2symbols(line); // prevod utf-8 na ascii
		index_s = 0;
		first_par = true; // prvni slovo?
		capital = false;
	}
	// nacitani znakupri
	if (!input->eof()) {
		b = a; // predchozi znak
		a = line[line_index]; // akt. znak
		line_index++;
		if (a == ';' && ulist.Move->next != NULL && line_index == ulist.Move->next->offset_id) { // je znak konec utf-8 znaku?
			// existuje dalsi utf-8 znak?
			ulist.Move = ulist.Move->next; // posun v seznamu na dalsi
		}

		if (!capital && a > 'A' && a < 'Z' && to_skip(b) && !to_skip(a)) {
			capital = true;
			//index_spell_start = index + 1;
			index_spell_start = line_index - 1;
		}

		if (quotes && a == '\"' && index_boundary < line_index-1) {
			//cout << b << a << " -quotes false !!! \n";
			quotes = false; // vypnutie blokovania vracania entit
			delimiter = false;
		}
		if (apostrophe && a == '\'' && index_boundary < line_index-1) {
			//cout << b << a << " -apostrophe false !!!\n";
			apostrophe = false; // vypnutie blokovania
			delimiter = false;
		}
		if (quotes || apostrophe) {
			index_next = line_index-1; // prekrytie backtrackingu
			index_s = index_next-1; // posuv indexu kvoli backtrackingu
		}

		if (delimiter && allow_boundary) {
			// pokusi sa najst druhe ohranicenie a blokovat vracanie entit
			// ktore nesiahaju od zaciatku po koniec ohranicenia uvodzovkami alebo apostrofom
			if (!apostrophe && !quotes && a == '\"')
				find_boundary(QUOTES);
			if (!quotes && !apostrophe && a == '\'' && to_skip(b))
				find_boundary(APOSTROPHE);
		}

		if (!apostrophe && !quotes && !to_skip(a) && to_skip(b) && first_par) { //
			index_next = line.find_first_of(delimiters, index_s + 1); // index pro
			first_par = false;                           // backtracking
		}

		if (index_next == std::string::npos) { // neexistuje dalsi slovo?
			index_next = 0; // reset
			first_par = true; // reset
		}
		index++;
	}
} // konec marker::next_char

//==========================================================//
// Nazev: next_word                                         //
// Trida: marker                                            //
// Popis: preskoci na zacatek nasledujiciho vyrazu v texte  //
//==========================================================//
void marker::next_word()
{
	index++;

	while ((to_skip(a)) && !input->eof()) {
		//cout << "Index: " << index_s << " ulist.move: " << ulist.Act->offset_id << " value: " << ulist.Act->char_value << endl;
		next_char(); // dalsi znak
		index_s++;
		index--;

		while (ulist.Act->next != NULL && to_skip(ulist.Act->next->char_value) && line_index == ulist.Act->next->offset_id - ulist.Act->next->count_sub) { // je treba se

			ulist.Act = ulist.Act->next;                  // posunout?
			ulist.Move = ulist.Act;                     // index utf-8

			if (ulist.Act->next == NULL) {             // znaku musi byt
				break;                                  // pred line_index
			}

			for (int i = ulist.Act->count_sub; i >= 0; i--) {
				next_char();
				index_s++;
			}

			index--;
		}
	}

	ulist.Move = ulist.Act; // nastaveni indexu na akt. polozku
	if (!begin) { // neni zacatek vyhledavani?
		index_s--;
	}

	begin = false;

	if (ulist.Act->next != NULL) { // existuje dalsi utf-8 znak?
		while (line_index > ulist.Act->next->offset_id) { // je treba se

			ulist.Act = ulist.Act->next;                  // posunout?
			ulist.Move = ulist.Act;                     // index utf-8

			if (ulist.Act->next == NULL) {             // znaku musi byt
				break;                                  // pred line_index
			}
		}
	}

	if (ulist.Act != NULL) { // aktualizovat index zacatku slova?
		first = index_global + index_s + 1 - ulist.Act->count_sub;
	}
} // konec marker::next_word

//======================================//
// Nazev: end_word                      //
// Trida: marker                        //
// Popis: skoci na konec slova v texte  //
//======================================//
void marker::end_word()
{

	if (apostrophe) {
		while ((a != '\'') && !input->eof()) {
			next_char(); // nacti dalsi znak
		}
	}
	else if (quotes) {
		while ((a != '\"') && !input->eof()) {
			next_char(); // nacti dalsi znak
		}
	}
	else {
		while (((!to_skip(a)) || a == '\\') && !input->eof()) {
			next_char(); // nacti dalsi znak
		}
	}

	if (((line.length() + 1) == line_index) && first_par) { // jsme na
		was_end = true;                               // konci radku?
	}
} // konec marker::end_word

//======================================================//
// Nazev: save_match                                    //
// Trida: marker                                        //
// Popis: ulozi nalezeny vyraz do promenne typu Tmatch  //
//======================================================//
Tmatch marker::save_match(std::string radek, std::string jmeno, unsigned long int f_index, unsigned long int l_index)
{
	Tmatch tmp;
	tmp.first_index = f_index; // index zacatku slova
	tmp.last_index = l_index; // index konce slova
	tmp.line_num = radek; // seznam cisel radku
	tmp.line_name = jmeno; // nazev entity
	return tmp;
} // konec marker::save_match

//=======================================//
// Nazev: lookup                         //
// Trida: marker                         //
// Popis: vyhledavani ve vstupnim texte  //
//=======================================//
std::string marker::lookup()
{
	dict_list *my_dict;
	fsa_arc_ptr *dummy = NULL;

	my_dict = &dictionary;
	ANNOT_SEPARATOR = my_dict->item()->annot_sep;
	set_dictionary(my_dict->item());

	potencial_match.first_index = 0;
	potencial_match.last_index = 0;
	index_next = 0;

	set_index(); // reset
	begin = true; // reset
	next_char(); // nacti prvni znak

	while (!input->eof()) { // je konec?
		next_word(); // dalsi slovo
		match_word(dummy->first_node(current_dict));

		if (spellchecking && capital && index_spell_start >= potencial_match.last_index) {
			if (spell_check(dummy->first_node(current_dict))) {
				if (index_next < index_spell_end) {
					index_next = index_spell_end;
				}
			}
		}

		// zavola se vyhledavaci funkce
		if (!quotes && !apostrophe && first_par && !was_end) { // prvni slovo?
			index_next = line.find_first_of(delimiters, index_s + 1);
		}
		// inkrementace a reset
		if (was_end)
			next_char();
		line_index = index_next + 1;
		index_s = index_next + 1;
		index = 0;
		first_par = true;
		a = ' ';

		if (was_end) {
			was_end = false;
			line_index = 0;
			index_s = 0;
		}
	}
	deleteItems(); // odstran vsechny polozky ze seznamu

	return dataset.results;
} // konec marker::lookup

//=====================================================================
// Nazev: auto_lookup
// Trida: marker
// Popis: vyhledavani ve vstupnim texte
//=====================================================================
std::string marker::auto_lookup(unsigned int many)
{
	dict_list *my_dict;
	fsa_arc_ptr *dummy = NULL;

	// premazanie stringu kvoli viacnasobnemu volaniu funkcie v jednej instancii triedy
	dataset.results.clear();
	my_dict = &dictionary;
	ANNOT_SEPARATOR = my_dict->item()->annot_sep;
	set_dictionary(my_dict->item());

	set_index();
	next_char();
	//while (!input->eof()) {
	next_word();
	auto_match_word(dummy->first_node(current_dict), many);
	//}

	return dataset.results;
} // konec marker::auto_lookup

//=======================================//
// Nazev: lookup_file                    //
// Trida: marker                         //
// Popis: vyhledavani ve vstupnim texte  //
//=======================================//
std::string marker::lookup_file(std::string input_name)
{
	set_input_filename(input_name);
	return lookup();
}

//=====================================================================
// Nazev: auto_lookup_file
// Trida: marker
// Popis: vyhledavani v souboru
//=====================================================================
std::string marker::auto_lookup_file(std::string input_name, unsigned int many)
{
	set_input_filename(input_name);
	return auto_lookup(many);
} // konec marker::auto_lookup_file

//=======================================//
// Nazev: lookup_string                  //
// Trida: marker                         //
// Popis: vyhledavani ve vstupnim texte  //
//=======================================//
std::string marker::lookup_string(std::string input_string) {
	std::stringstream ss;
	ss.str(input_string + "\n");

	set_input(&ss);
	return lookup();
}

//=====================================================================
// Nazev: auto_lookup_string
// Trida: marker
// Popis: vyhladavani ve vstupnim texte
//=====================================================================
std::string marker::auto_lookup_string(std::string input_string, unsigned int many)
{
	std::stringstream ss;
	ss.str(input_string);

	set_input(&ss);
	return auto_lookup(many);
} // konec marker::auto_lookup_string

//================================//
// Nazev: set_input_filename      //
// Trida: marker                  //
// Popis: otevre vystupni soubor  //
//================================//
void marker::set_input_filename(std::string name)
{
	if (dataset.input_file.is_open()) dataset.input_file.close();
	dataset.input_file.open(name.c_str());

	set_input(&dataset.input_file);
} // konec marker::set_input_filename


//========================================================//
// Nazev: set_input                                       //
// Trida: marker                                          //
// Parameter: input                                       //
// Popis: nastavuje vstup                                 //
//========================================================//
void marker::set_input(std::istream* in)
{
	input = in;
}

//================================================================//
// Nazev: match_word                                              //
// Trida: marker                                                  //
// Parameter: ukazovatel pro fsa, na zacatek automatu;            //
// Popis: prochazi znak po znaku automatem a hleda shodu          //
//================================================================//
void marker::match_word(fsa_arc_ptr start)
{
	bool found; // byla nalezena shoda znaku se zaznamem v automatu?
	bool fin = false; // je konec?
	bool save_mod = false; // je ulozen aspon jeden zaznam?
	fsa_arc_ptr test; // docasne ulozeny uzel
	fsa_arc_ptr next_node; // pracovni uzel
	std::string jmeno; // ukladani nazvu entity
	bool no_number = false; // fragment?

	do { // vyhledavani v KA
		next_node = start.set_next_node(current_dict);
		found = false;
		test = next_node;
		forallnodes(i) { // hledani shody znaku s automatem
			if (a == next_node.get_letter()) {
				jmeno.append(1, a);
				found = true; // znak byl nalezen
				start = next_node; // dalsi uzel
				break;
			}
		} // forallnodes....

		if (!found) {
			end_word();
			if (save_mod) {
				dataset.get_data(potencial_match, FIGA);
				save_mod = false;
				capital = false; // slovo zacinajuce velkym pismenom bolo rozpoznane
			}
			return;
		}

		// ak bol zadany prepinac "-o"
		if (entity_overlapping && to_skip(a) && !to_skip(b)) {	// vypis kvoli najdenemu oddelovaciemu znaku
			//end_word();
			if (save_mod) {  // ak bola rozpoznana entita
				dataset.get_data(potencial_match, FIGA);
				save_mod = false;
			}
		}

		// nacti dalsi znak
		next_char();

		test = start; // ulozeni uzlu
		next_node = start.set_next_node(current_dict);

		forallnodes(i) { // existuje zaznam v automatu
			if (('|' == next_node.get_letter()) && check_boundary() && (to_skip(a) || input->eof())) {
				// a != '&' je v podmienke kvôli vypisu entity ako casti slova ak za entitou nasleduje utf znak
				if (entity_overlapping || potencial_match.last_index < first || potencial_match.first_index == first) {
					// zakazanie vratenia prekryvajucej sa entity ak nebol zadany prepinac '-o'
					fin = true; // zacni s hledanim cisel radku
				}
				break;
			}
		}

		if (fin) { // byl najity zaznam v automatu, ted se bude rozpoznavat
			if (parse_line_numbers(&no_number, &save_mod, next_node))
				potencial_match = save_match(dataset.radek_line, jmeno, first, (first + index - 2) - (ulist.Move->count_sub - ulist.Act->count_sub));	// ulozeni info o entite do temp struktury
			else
				potencial_match = save_match(dataset.radek_line, jmeno, first, (first + index - 2));
			fin = false;
			capital = false; // slovo zacinajuce velkym pismenom bolo rozpoznane
		}

		if (fin && input->eof())	// vypis entity (ak sa nejaka nasla) kvoli koncu riadku
			dataset.get_data(potencial_match, FIGA); // uloz zaznam do stringu

		next_node = test; // navraceni se na puvodni uzel
		start = test;

	} while (!input->eof() && found);
} // konec marker::match_word

//=====================================================================
// Nazev: auto_match_word
// Trida: marker
// Parametry: 1) ukazatel pro fsa, na zacatek automatu;
//            2) kolik entit ma byt vypsano na vystup
// Popis: prochazi znak po znaku automatem a hleda shodu, pote dohledava
//        vsechny mozno pokracovani (max = kolik)
//=====================================================================
void marker::auto_match_word(fsa_arc_ptr start, int kolik)
{
	bool found = false; // byla nalezena shoda znaku se zaznamem v automatu?
	bool fin = false; // ma se zacit hledat potencialni zaznamy?
	bool done = false; // je dokonceno hledani potencialnich zaznamu?
	bool con_search = false; // maji se hledat cisla radku?
	bool go_back = false; // nasel se stejny zaznam s lehce jinym nazvem?
	const fsa_arc_ptr empty; // prazdny uzol pre ucely kopirovania
	int top = 1; // pocet poloziek vo vectoroch
	int many;

	std::vector <std::string> jmeno (1, ""); // ukladani nazvu entity
	std::vector <std::string> preradek (1, "");
	std::vector <fsa_arc_ptr> pole (1, empty); // uzel pro jednotlive zaznamy

	fsa_arc_ptr test; // docasne ulozeny uzel
	fsa_arc_ptr next_node; // pracovni uzel

	do { // vyhledavani v KA
		next_node = start.set_next_node(current_dict);
		found = false;

		test = next_node;

		forallnodes(i) { // hledani shody znaku s automatem
			if (a == next_node.get_letter()) { // shoda?
				start = next_node; // ulozeni uzlu s nalezenou shodou
				pole[0] = next_node;
				jmeno[0].append(1, a);
				found = true; // znak byl nalezen
				break;
			}
		} // forallnodes....

		if (!found && a == '\0' && (line.length() > 0)) { // nebyla nalezena shoda?
			end_word();
			fin = true; // zacni dohledavat
		}
		else if (!found) {
			end_word();
			return;
		}

		next_char();

		if (fin) { // bylo ukonceno vyhledavani v KA, zacina dohledavani potec. zaznamu

			fsa_arc_ptr temp;
			fsa_arc_ptr prvni;
			bool shot = false; // nasel shodu?
			bool end = false; // konec uzlu?
			bool first = true; // prvni uzel, ktery je testovan
			bool orig_node = false; // byl nalezen vysledek uz ve vyhledavani?
			bool no_number = false; // fragment?

			int act_cnt = 0; // aktualni vrchol (promenny)
			size_t indexik = 0; // index do abedy
			int cnt = 0; // aktualni vrchol (stabilni)
			std::string temp_name; // docasne ulozeni nazvu
			std::string abeda = " ',-.:aAbBcCdDeEfFgGhHiIjJkKlLmMnNoOpPqQrRsStTuUvVwWxXyYzZ&0123456789#;|"; // abeceda priority (pismena vice vlevo maji vyssi prioritu)
			size_t max = abeda.length(); // max znaku v abede
			next_node = pole[cnt].set_next_node(current_dict); // nastaveni na zacatek uzlu po vyhledavanem reztezci

			if (next_node.get_letter() == '|') { // pokud byl zaznam nalezen uz pri vyhledavani
				prvni = next_node; // ulozeni puvodniho uzlu
				orig_node = true; // byl nalezen ve vyhledavani
			}

			while ((return_all || cnt < kolik) && (cnt < top)) { // pobezi dokud budou polozky k dispozici
				do {
					temp = next_node; // ulozeni puvodniho uzlu
					do {
						many = 0;
						forallnodes(i) { // pro kazdy nasledujici uzel v akt. uzlu
							many++;
							if (abeda[indexik] == next_node.get_letter() || 0 == next_node.get_letter() || indexik >= max) { // je znak v uzlu?

								if (next_node.get_letter() == '|' || 0 == next_node.get_letter()) { // je znak konec uzlu?
									if (many > 1) {
										int pocet = 0;
										top++;	// pridanie polozky do vectorov
										jmeno.push_back("");
										preradek.push_back("");
										pole.push_back(empty);

										for (int x = cnt; x < (top - 1); x++) { // posunuti polozek dozadu ... aby se nestratil dosavadni posun
											pole[top - pocet - 1] = pole[top - pocet - 2];
											jmeno[top - pocet - 1] = jmeno[top - pocet - 2];
											pocet++;
										} // for
									} // if (many....

									shot = true; // nasel shodu
									end = true; // konec dohledavani
									test = next_node;
									break; // hura na vypis zaznamu do stringu
								} // if (next....

								if (indexik >= max) {		// nasleduje neznamy znak (nie je v abecede)
									shot = true;
									end = true; // konec dohledavani
									break;
								}

								temp_name = jmeno[cnt];

								jmeno[cnt].append(1, abeda[indexik]);	// ulozeni znaku do jmena (pokud to neni | )
								shot = true; // nasel shodu
								indexik = 0; // restart indexovani

								test = next_node; // ulozeni uzlu s nalezenym znakem
								first = true; //

								act_cnt = cnt; // nastaveni aktualniho vrcholu

								do { // smycka, ktera upresnuje zaznamy
									forallnodes(i) { // prochazeni vsech uzlu (sem se dostane pouze tehdy, pokud neni | )

										if (abeda[indexik] == next_node.get_letter()) { // nasel shodu v abede

											if (abeda[indexik] == '|') { // je znak konec uzlu?
												int pocet = 0;
												top++;	// pridanie polozky do vectorov
												jmeno.push_back("");
												preradek.push_back("");
												pole.push_back(empty);

												for (int x = cnt; x < (top - 1); x++) { // posunuti polozek dozadu ... aby se nestratil dosavadni posun
													pole[top - pocet - 1] = pole[top - pocet - 2];
													jmeno[top - pocet - 1] = jmeno[top - pocet - 2];
													pocet++;
												} // for...

												jmeno[cnt] = temp_name;
												pole[cnt] = next_node;
												shot = true; // nasel shodu
												end = true; // ?
												con_search = true;
												// test = next_node;
												break; // hura na vypis
											} // if (abeda...

											if (act_cnt > cnt) {
												int pocet = 0;
												top++;	// pridanie polozky do vectorov
												jmeno.push_back("");
												preradek.push_back("");
												pole.push_back(empty);

												for (int x = act_cnt; x < (top - 1); x++) { // pokud yy neni aktualni vrchol, pak se budou rotovat data
													pole[top - pocet - 1] = pole[top - pocet - 2]; // rotuje nody dozadu ... napr. 1 -> 2
													jmeno[top - pocet - 1] = jmeno[top - pocet - 2]; // rotuje nazev
													pocet++;
												}
											} // if (act....

											pole[act_cnt] = next_node; // ulozi node do aktualniho vrcholu

											if (act_cnt > cnt) {
												jmeno[act_cnt] = temp_name;
												jmeno[act_cnt].append(1, abeda[indexik]);
											} // if (act....

											act_cnt++; // zvysi aktualni vrchol
											break;
										}
									}

									indexik++; // posune se na dalsi znak v abecede priority
									next_node = temp; // vrati se na puvodni node, od ktereho se zacina

									if (indexik >= max) { // pokud se dosahlo max. indexu v abecede priority
										break; // konec
									}

								} while (/*(act_cnt < (kolik + 1)) && */(indexik < max)); // dokud bude vrchol v rozmezi a index do abecedy v rozmezi

								next_node = test.set_next_node(current_dict); // nastavi se dalsi pokrok v prvnim uzlu
								test = next_node; // ulozi se aktualni zacatek uzlu

								break;
							} // if (abeda....
						} // forallnodes....

						indexik++; // zvysi se index v abecede priority

						if (!first) { // prvni testovaci uzel?
							if (pole[cnt].arc == 0)
							{
								fprintf(stderr, "There is a zero in nodes-vector (id: pole) - something may be wrong.\n");
								return;
							}
							next_node = pole[cnt].set_next_node(current_dict);
							// nacte nasleduji uzel do akt. uzlu
						}
						else {
							next_node = test; // vrati zacatek uzlu do akt. uzlu
						}

					} while (!shot); // nasel shodu?

					shot = false; // reset
					indexik = 0; // reset
					next_node = test; // vrati zacatek uzlu do aktualniho uzlu

				} while (!end); // byl identifikovan konec uzlu?

				end = false; // reset
				first = false; // reset

				if (con_search) { // dohledavani skoncilo na zacatku uzlu
					next_node = pole[cnt];
				}

				if (orig_node) { // puvodni uzel, ktery byl nalezen uz ve vyhledavani
					next_node = prvni; // nacteni puvodniho uzlu do akt. uzlu
					orig_node = false; // reset
				}

				con_search = false; // reset
				if (next_node.get_letter() == '|') { // je zde info o cislo radku?
					con_search = true;
				}
				else {
					forallnodes(i)
						if (next_node.get_letter() == '|') { // je zde info o cislo radku
							con_search = true;
							break;
						}
				} // if (next....

				fin = false; // reset
				no_number = false; // reset

				if (con_search) { // nacitanie cisel riadkov z automatu
					if (parse_line_numbers(&no_number, NULL, next_node)) {
						potencial_match = save_match(dataset.radek_line, jmeno[cnt]);
						preradek[cnt] = dataset.radek_line;
					}
					else
						potencial_match = save_match(dataset.radek_line, jmeno[cnt]); // uloz info do zaznamu
					fin = true;
				} // if (con_s....
				con_search = false; // reset

				for (int i = 0; i < cnt; i++) {	// kontrola pred ulozenim
					if (dataset.radek_line == preradek[i]) {
						fin = false; // neulozi se do stringu
						go_back = true;
					}
					if (!strcmp(jmeno[i].c_str(), jmeno[cnt].c_str()))
					{
						fprintf(stderr, "Entities are repeating in output vector - something may be wrong.\n");
						return;  // zda se neopakuji zaznamy pokud ano, tak skonci
					}

				}

				if (no_number) {	// vyhodenie fragmentu z pola
						fin = false;  // neulozi sa do vypisoveho stringu
						go_back = true;	 // posunut entity
				}

				if (fin == true) {
					dataset.get_data(potencial_match, AUTOCOMPLETE); // nahrani zaznamu do vysledneho stringu
					fin = false; // reset
				}

				if (!go_back) {
					cnt++; // zvyseni aktualniho vrcholu polozek
				}
				else { // Posunuti entit (dalsi zaznam bude na stejnem miste)
					for (int x = cnt; x < (top - 1); x++) {
						pole[x] = pole[x + 1];
						jmeno[x] = jmeno[x + 1];
						preradek[x] = preradek[x + 1];
					}

					top--;
					pole.pop_back();
					jmeno.pop_back();
					preradek.pop_back();
				}
				go_back = false;
				if (cnt < top && pole[cnt].arc != 0)
					next_node = pole[cnt].set_next_node(current_dict); // nacteni uzlu dalsi polozky do akt. uzlu
			} // while (cnt....

			done = true; // dohledavani skoncilo
		} // if (fin....

	} while ((!input->eof() && found));

} // konec marker::auto_match_word

//=====================================================================
// Nazev: parse_line_numbers
// Trida: marker
// Parametry: 1) najdeny fragment
//            2) je ulozeny aspon jeden zaznam
//            3) uzol obsahujuci '|'
// Popis: nacita cisla riadkov z automatu a ulozi ich do premennej dataset.radek_line
// Ak je v save_mod hodnota NULL, metoda bola volana z autocomplete alebo zo spellcheckingu.
//=====================================================================
bool marker::parse_line_numbers(bool *no_number, bool *save_mod, fsa_arc_ptr next_node, std::vector<int> *line_numbers)
{
	dataset.radek_line.clear();                  // vsech radku v KB
	fsa_arc_ptr start = next_node;
	next_node = start.set_next_node(current_dict);
	start = next_node;
	unsigned int radek = 0; // temp pro hodnotu radku
	char znak = next_node.get_letter(); // prvni znak za '|'
	*no_number = false; // reset
	bool run = true; // ma zacit hledat?
	int bitshift = 21;      // bitovy posuv kvoli kodovaniu cisla v KA
	int byte_number = 4;    // cislo riadku je v KA zapisane na 4 bajtoch
	bool N_found = false;	// bol najdeny znak 'N' indikujuci fragment

	if (znak == '\n') { // chybny format?
		dataset.save_line(radek, true);
		if (save_mod)	// toto je kontrola spravnosti pointra
			*save_mod = true;
		run = false;
		return false;
	}

	while (run) { // ma se zacit vyhledavat?
		if ((byte_number > 0) && (znak != 'N')) {
			radek |= (((int)(znak & 0xFE)) << bitshift) >> 1;
			// odfiltrovanie spodneho bitu a priradenie bajtu do cisla
			bitshift -= 7;
			byte_number--;  // posun na dalsi byte
		}

		if (znak == 'N') { // ('N' = 78 in ASCII)
			radek = 0;	// reset
			byte_number = 4; // reset
			bitshift = 21; // reset
			N_found = true;
		}

		fsa_arc_ptr next_node = start.set_next_node(current_dict);
		start = next_node;
		znak = next_node.get_letter(); // dalsi znak v uzlu

		if ((byte_number == 0) && (znak % 2 == 0x01 || znak == 'N')) { // ak prave nacitany znak patri uz do dalsieho cisla alebo je to indikacia fragmentu ('N')
			dataset.save_line(radek, false);
			if (line_numbers) { // kontrola pointra
				line_numbers->push_back(radek); // pridanie do vectora cisel (spellchecking)
			}
			radek = 0; // reset
			byte_number = 4; // reset
			bitshift = 21; // reset
		}

		if ((znak == '\0')/* || (znak == '\n')*/) { // posledne cislo v riadku?      ('\0' = 0 ... '\n' = 10 in ASCII)
			if (!save_mod && N_found && !dataset.radek_line.empty())
				dataset.radek_line.pop_back();
			else
				dataset.save_line(radek, true);

			if (N_found && dataset.radek_line.empty()) // bol najdeny znak 'N' a ziadne cislo riadku
				*no_number = true;	// fragment entity

			if (line_numbers && radek) { // kontrola pointra
				line_numbers->push_back(radek); // pridanie do vectora cisel (spellchecking)
			}
			if (save_mod)	// kontrola pointra na NULL hodnotu
				*save_mod = true;
			return true;
		}
	}
}

//=====================================================================
// Nazev: find_boundary
// Trida: marker
// Parametry: 1) typ ohranicenia: uvodzovky/apostrofy
// Popis: hlada ohranicenie entit uvodzovkami alebo apostrofmi alebo kontroluje ohranicenie.
//=====================================================================
void marker::find_boundary(int type_of_boundary)
{
	int index_pair = 0;
	// "line_index" ukazuje na poziciu nasledujuceho znaku v case volania tejto funkcie
	if (type_of_boundary == QUOTES) {
		index_pair = line.find_first_of("\"", line_index);
		if (index_pair != std::string::npos && (index_pair - line_index-1 < 40)) {
			quotes = true;
			index_boundary = line_index-1;
			//std::cout << b << a << " -quotes true: " << line_index-1 << " - " << index_pair << std::endl;
			}
	}
	else if (type_of_boundary == APOSTROPHE) {
		index_pair = line.find_first_of("\'", line_index);
		if (index_pair != std::string::npos && (index_pair - line_index-1 < 40)) {
			apostrophe = true;
			index_boundary = line_index-1;
			//std::cout << b << a << " -apostrophe true: " << line_index-1 << " - " << index_pair << std::endl;
		}
	}
}

//=====================================================================
// Nazev: check_boundary
// Popis: Kontroluje ohranicenie entit uvodzovkami alebo apostrofmi.
//=====================================================================
bool marker::check_boundary()
{
	if (index_boundary < line_index-1 && ((quotes && a != '\"') || (apostrophe && a != '\'')))
		return false;
	else
		return true;
}

//==================================================================
// Nazev: spell_check
// Trida: marker
// Parameter: ukazovatel pro fsa, na zacatek automatu;
// Popis:  Tato metoda je spustana ak bolo najdene slovo zacinajuce
//         velkym pismenom a nebola v nom rozpoznana entita.
//       V parametroch vyzaduje iba korenovy uzol hlavneho automatu
//       ale takisto musi byt ulozene spravne cislo v premennej
//       index_spell_start - oznacujuce index daneho velkeho pismena
//       vo vete (v tomto pripade sa indexuje od 0)
// Navratova hodnota:   true - ak bol spellchecking uspesny, inak false
//==================================================================
bool marker::spell_check(fsa_arc_ptr start_node)
{
	std::vector<int> line_numbers; // cisla riadkov vsetkych moznych nahrad dokopy
	std::vector<std::string> replacements; // mozne nahrady slova
	word_list *repl_words;
	stringstream is;
	int first_index, last_index; // indexy prveho a posledneho pismena slova ktore bude opravovane

	// prechadzanie stringu a urcovanie konca slova, ktore sa bude opravovat
	char newch = line[index_spell_start], oldch = newch;
	for (int i = index_spell_start; i < line.size(); i++) {
		//oldch = newch;
		newch = line[i];
		if (to_skip(newch, i + 1) /*&& !to_skip(oldch, i+1)*/) {
			//cout << newch << " " << i+1 << endl << ulist.Move->offset_id << endl;
			index_spell_end = i - 1;
			break;
		}
		else if (i == line.size() - 1) { // ak vyhladavanie dosiahlo konca vety
			index_spell_end = i;
			break;
		}
	}

	// vynatie podretazca z vety
	string original = line.substr(index_spell_start, index_spell_end - index_spell_start + 1);
	is.str(original + "\n");
	//cout << "start: " << index_spell_start << "; end: " << index_spell_end << "; " << original << endl;
	tr_io io_obj(&is, std::cout, MAX_LINE_LEN, "", spell_dict->get_syntax());

	// force search for replacement candidates (the second argument)
	repl_words = spell_dict->spell_file(DISTANCE, false, io_obj);
	if (repl_words && repl_words->how_many()) {
		for (repl_words->reset(); repl_words->item(); repl_words->next()) {
			replacements.push_back(repl_words->item());
			//cout << repl_words->item() << endl;
		}

		// vrati cisla riadkov vsetkych entit ulozenych v replacements do line_numbers
		return_numbers(start_node, &line_numbers, &replacements);
		dataset.radek_line.clear(); // premazat stary zaznam, bude sa ukladat novy
		int list_size = line_numbers.size(); // pocet cisel v line_numbers

		// ukladanie cisel do stringu s oddelovacim znakom ';'
		for (int i = 0; i < list_size; i++) {
			if (i == list_size-1)
				dataset.save_line(line_numbers[i], true);
			else
				dataset.save_line(line_numbers[i], false);
		}

		// ak sa v opravovanom slove nachadza utf-8 znak - korekcia koncoveho offsetu
		TListptr *utf = ulist.Act;
		for (; utf->next != NULL; utf = utf->next) {
			if (utf->next->offset_id > index_spell_end + 1) {
				//cout << ulist.Act->count_sub << endl;
				//cout << utf->count_sub << endl;
				//cout << utf->next->offset_id << endl;
				//cout << index_spell_end << endl;
				break;
			}
		}

		// urcenie pociatocneho a koncoveho offsetu opravovaneho slova
		first_index = index_global + index_spell_start + 1 - ulist.Act->count_sub;
		last_index = first_index + index_spell_end - index_spell_start;
		last_index += ulist.Act->count_sub - utf->count_sub;
		//cout << first_index << endl << last_index << endl;

		// ukladanie informacii do docasneho zaznamu a vypis do vysledneho stringu
		potencial_match = save_match(dataset.radek_line, original, first_index, last_index);
		dataset.get_data(potencial_match, SPELLCHECKING);
		//dataset.results.append("spell\n\n");

		// vyprazdnenie listu s moznymi nahradami opravovaneho slova
		spell_dict->replacements.empty_list();
		capital = false; // reset
		return true; // spellchecking
	}

	spell_dict->replacements.empty_list();
	capital = false; // reset
	return false;

	//for (int i = 0; i < line_numbers.size(); i++)
		//cout << line_numbers[i] << endl;
}

//================================================================//
// Nazev: return_numbers                                          //
// Trida: marker                                                  //
// Parameter: ukazovatel pro fsa, na zacatek automatu;            //
//            ukazatel na vector cisel                            //
//            ukazatel na vector entit                            //
// Popis:  prejde vector entit a vrati vsetky cisla riadkov       //
//         tychto entit v jednom vectore                          //
//================================================================//
void marker::return_numbers(fsa_arc_ptr start, std::vector<int> *line_numbers, std::vector<std::string> *replacements)
{
	bool fin = false; // je konec?
	bool found; // byla nalezena shoda znaku se zaznamem v automatu?
	fsa_arc_ptr next_node; // pracovni uzel
	bool no_number = false; // fragment?
	int pos;
	fsa_arc_ptr head_node = start;

	for (int i = replacements->size()-1; i >= 0; i--) { // vyhledavani v KA
		//cout << (*replacements)[i] << endl;
		pos = 0;
		next_node = head_node.set_next_node(current_dict);
		found = false;

		for (int pos = 0; pos < (*replacements)[i].size(); pos++) {
			found = false;
			forallnodes(i) { // hledani shody znaku s automatem
				if ((*replacements)[i][pos] == next_node.get_letter()) {
					start = next_node; // dalsi uzel
					found = true;
					break;
				}
			} // forallnodes....
			next_node = start.set_next_node(current_dict);
		}


		if (found) {
			forallnodes(i) {
				// existuje zaznam v automatu
				if ('|' == next_node.get_letter()) {
					fin = true;
					break;
				}
			}
		}

		if (fin) { // byl najity zaznam v automatu, ted se bude rozpoznavat
			parse_line_numbers(&no_number, NULL, next_node, line_numbers);
			fin = false;
		}

		//for (int i = 0; i < line_numbers->size(); i++)
		//	cout << (*line_numbers)[i] << endl;

		replacements->pop_back();
	} // for
} // konec marker::return_numbers
