#!/usr/bin/python3

#AUTHOR - Radovan Sroka 1BIB, xsroka00@stud.fit.vutbr.cz
#LINK - https://knot.fit.vutbr.cz/wiki/index.php/Kb_filtering


import argparse, sys, re, os, unicodedata, name_filter #some imports
		
# some global variables fo parameter parsing
kbase = None
headkb = None
dest = None

entities = None
i_alias = None

delfile = None

def parse_params(): 

	"""this function parse parameters and return nothing, everything is in global varialbes"""
	
	global kbase
	global headkb
	global entities
	global dest	
	global delfile
	
	parser = argparse.ArgumentParser(description='Filtering KnowledgeBase')
	parser.add_argument('-k', help='KnowledgeBase path', dest='kbase')
	parser.add_argument('-t', help='Head-KB path', dest='headkb')
	parser.add_argument('-d', help='Destination file, default=KB-new(beside KB)', dest='dest')
	parser.add_argument('-e', nargs='+', help='Enumerate entities for filtering, default=all', default=["all"], dest="ent")
	parser.add_argument('-f', help='Save deleted names to file', dest='delfile', default=None)
																		
	args = parser.parse_args()
	
	kbase = args.kbase
	headkb = args.headkb
	entities = args.ent
	delfile = args.delfile
	
	if kbase is None or headkb is None:
		print ("I need -k and -t parameter")
		sys.exit(1)
	
	if args.dest is None:
		path = kbase.split("/")	
		path.pop()
		path.append('KB-new')
		dest = '/'.join(path)
	else:
		dest = args.dest	
	
	
		

def parse_head(): 

	"""this function parse HEAD-KB file for column position of important information"""
	
	head_file = open(headkb, "r")	
	lines = head_file.readlines()
	head_file.close()

	global i_alias
	i_alias = {}
	
	
	for line in lines:
		split_line = line.split("\t")
		if len(split_line) > 2:
			name = split_line[0][1:-3]
			name = name.split("_")
			name = " ".join(name).upper()
			for index in range(len(split_line)):
				#searching column like these
				if split_line[index] in {'{m}ALIAS', '{m}ALTERNATIVE NAME', '{m}OTHER TERM'}:
					i_alias[name] = index
					

def main():
	
	#initialization
	parse_params()
	parse_head()
	
	global i_alias
	global entities
	global dest
	global delfile
	
	for i in range(len(entities)):
		#set entities to upper
		entities[i] = entities[i].upper()
	
	#open files from parameters
	KB = open(kbase, "r")
	KB_new = open(dest, "w")
	if delfile is not None:
		DELFile= open(delfile, "w")
	
	filter_all = False
	if "ALL" in entities:
		filter_all = True
	
	for line in KB:	
		#reading KB
		split = line.split('\t')
		if len(split) < 5:
			continue
		type = split[1].upper()
		type = type.split("_")
		type = " ".join(type)
		alt_list = None
		
		if type not in i_alias:		
			#checking HEAD-KB
			KB.close()
			KB_new.close()
			if delfile is not None:
				DELFile.close()	
			print ("Type is " + type + " not in HEAD" + str(i_alias) )
			return 1
		
		
		if filter_all:
			#calling filter
			alt_list = split[i_alias[type]]	
		else:
			if type in entities:
				alt_list = split[i_alias[type]]
			else:
				continue
		alt_list_split = alt_list.split("|")
		
		alt = name_filter.check_alt_names(alt_list_split, split[2])
		
		if len(alt) > 0 and alt[0] is split[2]:
			alt.pop(0)
	
		#writing output
		split[i_alias[type]] = "|".join(alt)
		KB_new.write("\t".join(split))	
		
		if delfile is not None:
			sub = list(set(alt_list_split) - set(alt))
			DELFile.write("\t".join(split[:i_alias[type]]))		
			DELFile.write("\t" + split[i_alias[type]])
			DELFile.write("\t" + "|".join(sub))
			
			c_name = "OK"
			if name_filter.check_name(split[2]) is True:
				c_name = "BAD NAME"
			DELFile.write("\t" + c_name + "\n")	
			
			
	#closing files
	KB.close()
	KB_new.close()
	if delfile is not None:
		DELFile.close()
	
	return 0

main()
	
