#!/usr/bin/env python
# -*- coding: utf-8 -*-

#AUTHOR - Radovan Sroka 1BIB, xsroka00@stud.fit.vutbr.cz
#LINK - https://knot.fit.vutbr.cz/wiki/index.php/Kb_filtering


import re
import unicodedata		#some imports
import functools
import sys

def find_bigrams(input_string):	

	"""function decompose string from parameter to list of bigrams and return it ["name" => na-am-me]"""
	
	bigram_string = []
	for i in range(len(input_string)-1):
		bigram_string.append((input_string[i].lower() + input_string[i+1].lower()))
	return bigram_string

def compare_bigrams(bigram_a, bigram_b): 	

	"""function compare two lists of bigrams and return similarity between 0-1"""
	
	if len(bigram_b) is 0 and len(bigram_a) is 0: return 0

	if len(bigram_b) < len(bigram_a):
		temp = bigram_a
		bigram_a = bigram_b
		bigram_b = temp
	
	my_dict = {}
	for i in range(len(bigram_a)):
		
		for j in range(len(bigram_b)):
			if i in my_dict:
				break
			if j in my_dict.values():
				continue
			if bigram_a[i] == bigram_b[j]: 
				my_dict[i] = j
			
	#calculate incommon bigrams
	incommon = len(my_dict)
	
	result = (2.0 * incommon) / (len(bigram_a) + len(bigram_b))	
	return result



def strip_accents(string): 

	"""function remove accemt from unicode string"""

	return ''.join(c for c in unicodedata.normalize('NFD', string)
		if unicodedata.category(c) != 'Mn')


def check_regex(string): 	

	"""function check if string parameter is regular fo regex and return T/F"""

	#remove accent from string
	my_str = strip_accents(string)
	expresion =  "[\u00C0-\u02AF\u1E00-\u1FF4a-zß\’\`\'\-., \"]*(\d+rd|\d+st|\d+nd|\d+th)?[\u00C0-\u02AF\u1E00-\u1FF4a-zß\’\`\'\-., \"]*"
	regex = re.compile(expresion, re.IGNORECASE)
	if regex.match(my_str).group() is not None:
		if regex.match(my_str).group() is my_str: 
			return False

	return True

def check_name(name): 		

	"""function check name with check_regex"""
	my_name = name
	if not my_name:
		return my_name
	if sys.version_info < (3,0):
		my_name = unicode(my_name)
	return check_regex(my_name)


def filter_regex(input_list): 	
	
	"""function filter iregular names from list"""

	return (i for i in input_list if not check_regex(i))		

def check_alt_names(input_list, name=None):

	"""
	function filter iregular names from list and control strings similarity
	parameters are list of alternative names and name of (person|artist.....)
	returns list of regular alternative names
	"""

	if not input_list:
		return None

	min_names = 5	#minimum alternative names for similarity control
	sensitivity = 0.1	#set sensitivity
	reg = True	#set regex control

	ratio = {}
	bigrams = {}
	delete = []
	my_list = input_list[:]
	
	if sys.version_info < (3,0):
		my_list = list(map(lambda x: unicode(x), my_list))

	my_name = name
	name_type = None
	if not name:
		if sys.version_info < (3,0):
			my_name = unicode("")
		else:
			my_name = ""
		name_type = type(my_name)
	else:
		name_type = type(my_name)
		if sys.version_info < (3,0):
			my_name = unicode(my_name)
		my_list = [my_name] + my_list

	entity_bi = find_bigrams(strip_accents(my_name))

	if reg: 
		my_list = list(filter_regex(my_list))

	if len(my_list) < int(min_names) or int(min_names) is -1: return my_list
	if len(my_list) <= 1: return my_list

	for item in my_list:
		ratio[item] = 0
		bigrams[item] = find_bigrams(strip_accents(item)) 

	for base in my_list:	
		#compute ratio between names
		for actual in my_list:
			if base is actual: continue
			ratio[base] += compare_bigrams(bigrams[base], bigrams[actual])

	for key in ratio:
		#compute average ratio
		ratio[key] /= len(ratio)

		#compute ratio with name
		ratio_with_name = compare_bigrams(bigrams[key], entity_bi)

		if ratio_with_name >= 0.5:
			continue

		if ratio[key] < sensitivity:
			delete.append(key)
			continue

	for item in delete:
		my_list.remove(item)

	if sys.version_info < (3,0):
		if name_type is str:
			my_list = list(map(lambda x: str(x), my_list))
	
	return my_list

