#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Brno University of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Author: Matej Magdolen, xmagdo00@stud.fit.vutbr.cz
# Author: Jan Doležal, xdolez52@stud.fit.vutbr.cz
# Author: Lubomír Otrusina, iotrusina@fit.vutbr.cz
#
# Description: Loads a shared knowledge base.

import os, sys
import re
import imp
import cPickle as pickle
import threading
import subprocess
import unicodedata

# Pro debugování:
import inspect
DEBUG_EN = True
DEBUG_EN = False

# KONSTANTY:

# Získání absolutní cesty k adresáři ve kterém je tento soubor.
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
DIRPATH_KB_DAEMON = os.path.abspath(os.path.join(SCRIPT_DIR, "../SharedKB/var2"))
PATH_KB_DAEMON = os.path.abspath(os.path.join(DIRPATH_KB_DAEMON, "decipherKB-daemon"))
PATH_KB = os.path.abspath(os.path.join(SCRIPT_DIR, "../KB-HEAD.all"))

KB_MULTIVALUE_DELIM = "|"

# # Timeouty v sekundách:
Timeout_SharedKB_start = 300
Timeout_process_exists = 10

reload(sys)
sys.setdefaultencoding("utf-8")

# FUNKCE A TŘÍDY:

def remove_accent(_string):
    """Removes accents from a string. For example, Eduard Ovčáček -> Eduard Ovcacek."""

    nkfd_form = unicodedata.normalize('NFKD', unicode(_string))
    return str("".join([c for c in nkfd_form if not unicodedata.combining(c)]))

def update_globals(kbd_dir, kb_path, kb_delim = '|', timeout_s = 300, timeout_e = 10):
	global DIRPATH_KB_DAEMON, PATH_KB_DAEMON, PATH_KB, KB_MULTIVALUE_DELIM, Timeout_SharedKB_start, Timeout_process_exists
	DIRPATH_KB_DAEMON = kbd_dir
	PATH_KB_DAEMON = os.path.join(DIRPATH_KB_DAEMON, "decipherKB-daemon")
	PATH_KB = kb_path

class KnowledgeBase:
	'''
	Třída zapouzdřující KB.
	'''

	def __init__(self, kb_shm_name=None):
		'''
		Inicializace.
		'''

		KB_shm = imp.load_source('KB_shm', os.path.join(DIRPATH_KB_DAEMON,"KB_shm.py"))
		self.kb_shm = KB_shm.KB_shm(kb_shm_name)
		self.subKB_ps = None
	
	def start(self):
		'''
		Připojí sdílenou paměť.
		'''

		KB_run = KB_is_running()
		self.subKB_ps = None
		
		try:
			if not KB_run:
				self.subKB_ps = run_subKB()
			self.kb_shm.start()
		except:
			if self.subKB_ps != None:
				exit_subKB(self.subKB_ps)
			self.kb_shm.end()
			self.subKB_ps = None
			raise
	
	def end(self):
		'''
		Odpojí sdílenou paměť.
		'''

		if self.subKB_ps != None:
			exit_subKB(self.subKB_ps)
		self.kb_shm.end()
		self.subKB_ps = None
	
	def initName_dict(self):
		'''
		Dictionary asociates parts of person names with corresponding items of knowledge base.
		'''

		PATH_NAMEDICT = os.path.join(SCRIPT_DIR, "ner_namedict.pkl")
		
		person_alias = self.get_head_col("person", "ALIAS")
		person_name = self.get_head_col("person", "NAME")
		artist_other = self.get_head_col("artist", "OTHER TERM")
		artist_preferred = self.get_head_col("artist", "PREFERRED TERM")
		artist_display = self.get_head_col("artist", "DISPLAY TERM")
		
		self.name_dict = {}
		self.fragments = set()
		
		# Proto aby se nemusela znova procházet KB, vytvoří se soubor PATH_NAMEDICT.
		# Namedict se bude načítat z něj pokud PATH_KB bude starší než PATH_NAMEDICT - tím dojde k urychlení.
		if (os.access(PATH_NAMEDICT, os.F_OK)) and (os.stat(PATH_KB).st_mtime < os.stat(PATH_NAMEDICT).st_mtime):
			namedict_file = open(PATH_NAMEDICT, 'rb')
			self.name_dict = pickle.load(namedict_file)
			namedict_file.close()
		else:
			namedict_file = open(PATH_NAMEDICT, 'wb')
			line = 1
			text = self.get_data_at(line, 1)
			while text != None:
				ent_type = self.get_ent_type(line)
				if ent_type in ["person", "artist"]:
					if ent_type == "person":
						whole_names = self.get_data_at(line, person_alias).split(KB_MULTIVALUE_DELIM)
						whole_names.append(self.get_data_at(line, person_name))
					elif ent_type == "artist":
						whole_names = self.get_data_at(line, artist_other).split(KB_MULTIVALUE_DELIM)
						whole_names.append(self.get_data_at(line, artist_preferred))
						whole_names.append(self.get_data_at(line, artist_display))

					# creates subnames
					names = self.get_subnames(whole_names, ent_type, line)

					for name in names:
						name = remove_accent(name).lower()
						if name not in self.name_dict:
							self.name_dict[name] = set([line])
						else:
							self.name_dict[name].add(line)
				line += 1
				text = self.get_data_at(line, 1)
			pickle.dump(self.name_dict, namedict_file, pickle.HIGHEST_PROTOCOL)
			namedict_file.close()
	
	def get_subnames(self, whole_names, ent_type, line):
		'''
		From a list of whole names for a given person, it creates a set of all possible subnames.
		For example, for the name name "George Washington", it creates a set containing two subnames - "George" and "Washington".
		'''
		
		forbidden = ["Mr", "Sir", "Sr", "Jr", "The", "St", "University", "Saint"]
		roles = ["Baron", "Prince", "Duke", "Earl", "King", "Pope", "Queen", "Artist", "Painter"]
		regex_place = re.compile(" of .*")
		regex_role = re.compile("[Tt]he [a-zA-Z]+")
		regex_name = re.compile("[A-Z][a-z-']+[a-zA-Z]*[a-z]+") # this should match only a nice name
		names = set()
		roles = set()

		# getting roles for artists and persons
		if ent_type == "artist":
			roles.add("artist")
			preferred_role = self.get_data_for(line, "PREFERRED ROLE")
			if preferred_role and " " not in preferred_role:
				roles.add(preferred_role)
			other_roles = self.get_data_for(line, "OTHER ROLE").split("|")
			for other_role in other_roles:
				if other_role and " " not in other_role:
					roles.add(other_role)
		elif ent_type == "person":
			professions = self.get_data_for(line, "PROFESSION").split("|")
			for profession in professions:
				if profession:
					roles.add(profession)
		for role in roles:
			role = "the " + role.lower()
			names.add(role)
			self.fragments.add(role)
			self.fragments.add(role.title())
			self.fragments.add(role.replace("the", "The"))

		for whole_name in whole_names:
			# removing a part of the name with location information (e.g. " of Polestown" from the name "Richard Butler of Polestown")
			whole_name = regex_place.sub("", whole_name)

			# searching for a role
			role = regex_role.match(whole_name)
			if role:
				match = role.group()
				if match.split(" ")[1] in roles:
					names.add(match)
					match = match.lower()
					self.fragments.add(match)
					self.fragments.add(match.title())
					self.fragments.add(match.replace("the", "The"))

			# splitting the name to the parts
			subnames = whole_name.split(' ')
			for subname in subnames:
				if subname.endswith(","):
					subname = subname[:-1]
				# removing accent, because python re module doesn't support [A-Z] for Unicode
				subname_without_accent = remove_accent(subname)
				result = regex_name.match(subname_without_accent)
				if result:
					match = result.group()
					if match == subname_without_accent and match not in forbidden and match not in roles:
						names.add(subname)
						self.fragments.add(subname)
		return names
	
	def print_subnames(self):
		'''
		Print all partial name variants from self.name_dict.
		'''
		
		for fragment in self.fragments:
			print fragment

	def get_field(self, line, column):
		'''
		Zavrhovaná metoda!
		Číslování řádků od 1 a sloupců od 0, podle "ner.py".
		Ovšem SharedKB čísluje řádky i sloupce od 1.
		'''
		
		return self.kb_shm.dataAt(line, column + 1)
	
	def get_data_at(self, line, col):
		'''
		Číslování řádků i sloupců od 1.
		'''
		
		return self.kb_shm.dataAt(line, col)
	
	def get_data_for(self, line, col_name):
		'''
		Číslování řádků od 1.
		'''
		
		return self.kb_shm.dataFor(line, col_name)
	
	def get_head_at(self, line, col):
		'''
		Číslování řádků i sloupců od 1.
		'''
		
		return self.kb_shm.headAt(line, col)
	
	def get_head_for(self, ent_type, col):
		'''
		Číslování sloupců od 1.
		'''
		
		return self.kb_shm.headFor(ent_type, col)
	
	def get_head_col(self, ent_type, col_name):
		'''
		Vrátí číslo sloupce pro požadovaný ent_type a jméno sloupce.
		'''
		
		return self.kb_shm.headCol(ent_type, col_name)
	
	def get_complete_data(self, line, delim='\t'):
		'''
		Vrátí tuple (počet sloupců, celý řádek), kde v jednom řetězci je celý řádek pro požadovaný line, tak jak je v KB.
		Parametr delim umožňuje změnit oddělovač sloupců.
		'''
		
		text_line = ""  
		col = 1
		text = self.get_data_at(line, col)
		
		if text != None:
			text_line += text
			col += 1  
			text = self.get_data_at(line, col)
		
		while text != None:
			text_line += delim
			text_line += text
			col += 1
			text = self.get_data_at(line, col)
		
		return (col-1, text_line)
	
	def get_complete_head(self, ent_type, delim='\t'):
		'''
		Vrátí tuple (počet sloupců, celý řádek), kde v jednom řetězci je celý řádek pro požadovaný line, tak jak je v KB.
		Parametr delim umožňuje změnit oddělovač sloupců.
		'''
		
		text_line = ""  
		col = 1
		text = self.get_head_for(ent_type, col)
		
		if text != None:
			text_line += text
			col += 1  
			text = self.get_head_for(ent_type, col)
		
		while text != None:
			text_line += delim
			text_line += text
			col += 1
			text = self.get_head_for(ent_type, col)
		
		return (col-1, text_line)
	
	def get_complete_ent_pretty(self, line):
		'''
		Vrátí tuple (počet sloupců, celý řádek), kde v jednom řetězci je celý řádek pro požadovaný line, tak jak je v KB.
		Parametr delim umožňuje změnit oddělovač sloupců.
		'''
		
		text_line = ""  
		col = 1
		ent_type = self.get_ent_type(line)
		text_head = self.get_head_for(ent_type, col)
		text_data = self.get_data_at(line, col)
		
		if (text_head != None) and (text_data != None):
			text_line += text_head + ": " + text_data
			col += 1
			text_head = self.get_head_for(ent_type, col)
			text_data = self.get_data_at(line, col)
		
		while (text_head != None) and (text_data != None):
			text_line += "\n"
			text_line += text_head + ": " + text_data
			col += 1
			text_head = self.get_head_for(ent_type, col)
			text_data = self.get_data_at(line, col)
		
		return (col - 1, text_line)
	
	def get_ent_type(self, line):
		"""Returns a type of an entity at the line of the knowledge base"""
		
		return self.kb_shm.dataType(line)
	
	def get_dates(self, line):
		ent_type = self.get_ent_type(line)
		if ent_type in ["person", "artist"]:
			dates = set([self.get_data_for(line, "DATE OF BIRTH"), self.get_data_for(line, "DATE OF DEATH")])
			dates.discard("")
			return dates
		return set()
	
	def get_location_code(self, line):
		return self.get_data_for(line, "FEATURE CODE")[0:3]
	
	def get_nationalities(self, line):
		ent_type = self.get_ent_type(line)
		if ent_type == "nationality":
			nation = self.get_data_for(line, "ALIAS").split(KB_MULTIVALUE_DELIM)
			nation.extend(self.get_data_for(line, "ADJECTIVAL FORM").split(KB_MULTIVALUE_DELIM))
			nation.append(self.get_data_for(line, "NAME"))
			nation.append(self.get_data_for(line, "COUNTRY NAME"))
		elif ent_type == "person":
			nation = self.get_data_for(line, "NATIONALITY").split(KB_MULTIVALUE_DELIM)
		elif ent_type == "artist":
			nation = self.get_data_for(line, "OTHER NATIONALITY").split(KB_MULTIVALUE_DELIM)
			nation.append(self.get_data_for(line, "PREFERRED NATIONALITY"))
		nation = set([nat.lower() for nat in nation if nat != ""])
		return nation
	
	def get_score(self, line):
		"""
		Returns disambiguation score based on Wikipedia statistics and score based on other metrics.
		"""

		result = self.get_data_for(line, "CONFIDENCE")
		
		try:
			return float(result)
		except:
			err_type = self.get_ent_type(line)
			if err_type == None:
				err_head = (None, None)
				err_data = (None, None)
			else:
				err_head = self.get_complete_head(err_type)
				err_data = self.get_complete_data(line)
			sys.stderr.write(cur_inspect() + ": Line \"" + str(line) + "\" have non-integer type \"" + str(type(result)) + "\" with content \"" + str(result) + "\"\n")
			sys.stderr.write(cur_inspect() + ": Dump head for type \"" + str(err_type) + "\" (cols=" + str(err_head[0]) + "):\n\"\"\"\n" + str(err_head[1]) + "\n\"\"\"\n")
			sys.stderr.write(cur_inspect() + ": Dump line \"" + str(line) + "\" (cols=" + str(err_data[0]) + "):\n\"\"\"\n" + str(err_data[1]) + "\n\"\"\"\n")
			raise
	
	def people_named(self, subname):
		"""
		Returns all names (KB ids) containing a given subname.
		"""

		if subname in self.name_dict:
			return self.name_dict[subname]
		else:
			return set()

class TimeoutError(Exception):
	pass

def print_dbg(text=""):
	if not DEBUG_EN:
		return
	callerframerecord = inspect.stack()[1]
	frame = callerframerecord[0]
	info = inspect.getframeinfo(frame)
	
	head = "(" + info.filename + ", " + info.function + ", " + str(info.lineno) + ")"
	
	print( head + ":\n'''\n" + text + "\n'''" )

def cur_inspect():
	callerframerecord = inspect.stack()[1]
	frame = callerframerecord[0]
	info = inspect.getframeinfo(frame)
	
	head = "(" + info.filename + ", " + info.function + ", " + str(info.lineno) + ")"
	
	return head

def exec_function(func, args=(), kwargs={}, timeout_duration=10, default=None):
	'''
	This function will spawn a thread and run the given function
	using the args, kwargs and return the given default value if the
	timeout_duration is exceeded.
	'''

	class InterruptableThread(threading.Thread):
		def __init__(self):
			threading.Thread.__init__(self)
			self.result = default
		def run(self):
			self.result = func(*args, **kwargs)
	it = InterruptableThread()
	it.start()
	it.join(timeout_duration)
	if it.isAlive():
		return it.result
	else:
		return it.result

def subprocess_read(ps, ps_output_stream, timeout=None):
	'''
	ps - subproces
	ps_output_stream - výstupní stream subprocesu
	timeout - maximální doba čtení [sekundy]
	'''

	if (ps.poll() != None):
		raise Exception("Internal Error: " + cur_inspect())
	
	result = exec_function(ps_output_stream.read, timeout_duration=timeout)
	if (result == None):
		ps.terminate()
		ps.wait()
		raise TimeoutError
	
	return result

def subprocess_readline(ps, ps_output_stream, timeout=None):
	'''
	ps - subproces
	ps_output_stream - výstupní stream subprocesu
	timeout - maximální doba čtení [sekundy]
	'''

	if (ps.poll() != None):
		raise Exception("Internal Error: " + cur_inspect())
	
	result = exec_function(ps_output_stream.readline, timeout_duration=timeout)
	if (result == None):
		ps.terminate()
		ps.wait()
		raise TimeoutError
	
	return result

def process_exists(proc, id = False):
	'''
	proc       -> name/id of the process
	id = True  -> search for pid
	id = False -> search for name by regex (default)
	'''

	CMD = "ps -e -o pid= -o cmd="
	
	ps = subprocess.Popen(CMD, shell=True, stdout=subprocess.PIPE)
	try:
		output = subprocess_read(ps, ps.stdout, timeout=Timeout_process_exists)
	except TimeoutError:
		raise RuntimeError("Timeout of subprocess \"" + CMD + "\".")
	ps.stdout.close()
	ps.wait()
	
	for line in output.split("\n"):
		if line != "" and line != None:
			fields = line.split()
			pid = fields[0]
			pname = fields[1]
		
		if (not id):
			if re.search("^(?:.*/)?" + proc + "(?:\s+.*)?$", pname):
				return True
		else:
			if (proc == pid):
				return True
	return False

def KB_is_running():
	if process_exists(os.path.basename(PATH_KB_DAEMON)):
		return True
	return False

def run_subKB():
	ps = subprocess.Popen([PATH_KB_DAEMON, PATH_KB], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	try:
		output = subprocess_readline(ps, ps.stdout, timeout=Timeout_SharedKB_start)
	except TimeoutError:
		raise RuntimeError("Timeout of subprocess \"" + PATH_KB_DAEMON + "\".")
	
	print_dbg( output )
	
	if (ps.poll() != None):
		ps_stdout = ps.stdout.read()
		ps_stderr = ps.stderr.read()
		if (ps_stdout):
			sys.stderr.write(PATH_KB_DAEMON + " [STDOUT]:\n")
			sys.stderr.write(ps_stdout+"\n")
		if (ps_stderr):
			sys.stderr.write(PATH_KB_DAEMON + " [STDERR]:\n")
			sys.stderr.write(ps_stderr+"\n")
		
		ps.stdout.close()
		ps.stderr.close()
		ps.wait()
		ps = None
		raise RuntimeError(PATH_KB_DAEMON + " has failed to start")
	
	return ps

def exit_subKB(ps):
	ps.terminate()
	if (ps.wait() != 0):
		ps_stdout = ps.stdout.read()
		ps_stderr = ps.stderr.read()
		if (ps_stdout):
			sys.stderr.write(PATH_KB_DAEMON + " [STDOUT]:\n")
			sys.stderr.write(ps_stdout+"\n")
		if (ps_stderr):
			sys.stderr.write(PATH_KB_DAEMON + " [STDERR]:\n")
			sys.stderr.write(ps_stderr+"\n")
		
	ps.stdout.close()
	ps.stderr.close()
	return ps.wait()

# konec souboru ner_knowledge_base.py
