# -*- coding: utf-8 -*-
import os
import os.path
import sys
import urllib.request
import shutil
import re
from iritouri import iri2uri
from time import time, sleep

SAVE_FOLDER = "/mnt/athena3/kb/images/wikimedia"
RATE_LIMIT = float(1024)

class RateLimit(object):
	"""Rate limit a url fetch"""
	def __init__(self):
		self.start = time()
	def __call__(self, block_count, block_size, total_size):
		total_kb = total_size / 1024
		downloaded_kb = (block_count * block_size) / 1024
		elapsed_time = time() - self.start
		if elapsed_time != 0:
			rate = downloaded_kb / elapsed_time
			expected_time = downloaded_kb / RATE_LIMIT
			sleep_time = expected_time - elapsed_time
			if sleep_time > 0:
				sleep(sleep_time)

failed = 0
succ = 0

for line in sys.stdin:
	url = line.strip()
	if (len(line.strip().split("wikipedia/", 1)) > 1):
		dest = SAVE_FOLDER+"/"+line.strip().split("wikipedia/", 1)[1]
		if not os.path.isfile(dest):
			try:
				os.makedirs(re.sub('(.*)\/[^/]*', '\\1', dest), 0o777, True)
			except OSError as exc:
				if os.path.isdir(re.sub('(.*)\/[^/]*', '\\1', dest)):
					pass
				else: raise
			try:
				urllib.request.urlretrieve(iri2uri(url), dest, RateLimit())
				succ += 1
			except urllib.error.HTTPError as exc:
				sys.stdout.write(url+"\n")
				failed += 1
				pass
			if ((failed + succ) % 10 == 0):
				sys.stderr.write("Downloading images: Downloading in progress. Successfuly downloaded: "+str(succ)+"; Failed: "+str(failed)+".\n")
				sys.stderr.flush()

sys.stderr.write("Downloading images: Downloading in progress. Successfuly downloaded: "+str(succ)+"; Failed: "+str(failed)+".\n")
sys.stderr.write("Downloading images: Done.\n")